# -*- coding: utf-8 -*-
from adbutils.exceptions import AdbBaseError


class AdbProcessNotFound(AdbBaseError):
    """ An error while dumpsys meminfo process not found """


class AdbNoInfoReturn(AdbBaseError):
    """ An error while adb shell did not return the correct result """



# -*- coding: utf-8 -*-
from adbutils import ADBDevice
from adbutils.constant import AAPT_LOCAL_PATH, AAPT_REMOTE_PATH, ANDROID_TMP_PATH
from adbutils.exceptions import AdbBaseError

from typing import Union, Tuple, List, Optional, Match, Dict
import re
import time


class Aapt(object):
    def __init__(self, device: ADBDevice):
        self.device = device
        self._install_aapt()

    def get_app_info(self, name: str) -> Dict[str, str]:
        """
        获取app信息

        Args:
            name: app包名

        Returns:
            app信息, 包含:package_name/versionCode/versionName/sdkVersion/targetSdkVersion/app_name/launchable_activity
        """
        app_info = self._get_app_info(name)
        if m := self._parse_app_info(app_info):
            try:
                app_name = m.group('app_name_zh')
            except IndexError:
                app_name = m.group('app_name')

            return {
                'package_name': m.group('package_name'),
                'versionCode': m.group('versionCode'),
                'versionName': m.group('versionName'),
                'sdkVersion': m.group('sdkVersion'),
                'targetSdkVersion': m.group('targetSdkVersion'),
                'app_name': app_name,
                'launchable_activity': m.group('launchable_activity'),
            }

    @staticmethod
    def _parse_app_info(app_info: str) -> Optional[Match[str]]:
        """
        解析aapt d badging获取到的应用信息

        Args:
            app_info: aapt获取到的应用信息

        Returns:

        """
        pattern = re.compile(r'package: name=\'(?P<package_name>\S*)\' '
                             r'versionCode=\'(?P<versionCode>\S*)\' '
                             r'versionName=\'(?P<versionName>\S*)\''
                             r'.*sdkVersion:\'(?P<sdkVersion>\d+)\''
                             r'.*targetSdkVersion:\'(?P<targetSdkVersion>\d+)\''
                             r'.*application-label-zh:\'(?P<app_name_zh>\S+)\''
                             r'.*application: label=\'(?P<app_name>\S*)\''
                             r'.*launchable-activity: name=\'(?P<launchable_activity>\S*)\'', re.DOTALL)
        m = pattern.search(app_info)
        return m if m else None

    def _get_app_info(self, name: str) -> Optional[str]:
        app_path = self.device.get_app_install_path(name)
        # badging: Print the label and icon for the app declared in APK.
        if not app_path:
            raise AdbBaseError(f"'{name}' install path not found")

        ret = self.device.shell(f'{AAPT_REMOTE_PATH} d badging {app_path}')
        return ret

    def _get_app_path_list(self, flag_options: Union[None, str, list] = None) -> List[Tuple[str, str]]:
        """
        获取app的对应地址

        Args:
            flag_options: 获取app_list可指定参数,与pm list packages相同

        Returns:
            app_path_list: tuple[app_path, packageName]
        """
        options = ['-f']
        if isinstance(flag_options, list):
            if '-f' not in flag_options:
                options += flag_options
        elif isinstance(flag_options, str):
            options += [flag_options]
        app_list = self.device.app_list(options)
        pattern = re.compile('^(\\S+)=(\\S+)$')
        ret = []
        for app in app_list:
            m = pattern.findall(app)
            if m:
                ret.append(m[0])

        return ret

    def _install_aapt(self) -> None:
        """
        check if aapt installed

        Returns:
            None
        """
        if not self.device.check_file(ANDROID_TMP_PATH, 'aapt'):
            if 'arm' in self.device.abi_version:
                self.device.push(local=AAPT_LOCAL_PATH['arm'], remote=AAPT_REMOTE_PATH)
            else:
                self.device.push(local=AAPT_LOCAL_PATH['x86'], remote=AAPT_REMOTE_PATH)
            time.sleep(1)
            self.device.shell(['chmod', '755', AAPT_REMOTE_PATH])



# -*- coding: utf-8 -*-
import re
from typing import Match, Optional, Union, Dict, Tuple, List

from adbutils import ADBDevice
from .exceptions import AdbProcessNotFound


class Meminfo(object):
    def __init__(self, device: ADBDevice):
        """
        获取设备性能数据

        Args:
            device: adb设备类
        """
        self.device = device

    def get_pss_by_process(self, meminfo: Optional[str] = None) -> List[Tuple[int, str, int]]:
        """
        获取系统所有进程的pss内存大小,单位(KB)\n
        返回列表,每个参数都是tuple(memory,package_name,pid)

        Returns:
            所有进程的pss内存
        """
        meminfo = meminfo or self.get_system_meminfo()
        if m := self._parse_system_meminfo(meminfo):
            ret = []
            pss_by_process = m.group('pss_by_process').strip()
            pss_by_process = pss_by_process.splitlines()
            pattern = re.compile(r'(?P<memory>\S+K): (?P<name>\S+)\s*\(pid (?P<pid>\d+)')
            for process in pss_by_process:
                m = pattern.search(process.strip())
                memory, name, pid = self._pares_memory(m.group('memory')), m.group('name'), int(m.group('pid'))
                ret.append((memory, name, pid))
            return ret

    def get_total_ram(self, meminfo: Optional[str] = None) -> int:
        """
        获取全部内存(total_ram)大小,单位(KB)

        Args:
            meminfo: 内存信息

        Returns:
            内存大小(KB)
        """
        meminfo = meminfo or self.get_system_meminfo()
        if m := self._parse_system_meminfo(meminfo):
            return self._pares_memory(m.group('total_ram'))

    def get_free_ram(self, meminfo: Optional[str] = None) -> int:
        """
        获取Free RAM大小,单位(KB)

        Args:
            meminfo: 内存信息

        Returns:
            Free RAM(KB)
        """
        meminfo = meminfo or self.get_system_meminfo()
        if m := self._parse_system_meminfo(meminfo):
            return self._pares_memory(m.group('free_ram'))

    def get_used_ram(self, meminfo: Optional[str] = None) -> int:
        """
        获取Used RAM大小,单位(KB)

        Args:
            meminfo: 内存信息

        Returns:
            Used RAM(KB)
        """
        meminfo = meminfo or self.get_system_meminfo()
        if m := self._parse_system_meminfo(meminfo):
            return self._pares_memory(m.group('used_ram'))

    def get_lost_ram(self, meminfo: Optional[str] = None) -> int:
        """
        获取Lost RAM大小,单位(KB)

        Args:
            meminfo: 内存信息

        Returns:
            Lost RAM(KB)
        """
        meminfo = meminfo or self.get_system_meminfo()
        if m := self._parse_system_meminfo(meminfo):
            return self._pares_memory(m.group('lost_ram'))

    def get_app_meminfo(self, package: Union[str, int]) -> Optional[Dict[str, Dict[str, int]]]:
        """
         获取指定包或进程号的内存信息
         键为内存条目名称
         值是一个包含pss_total/private_dirty/private_clean/swapPss_dirty/heap_size/heap_alloc/heap_free的字典,对应的内存占用(KB)

        Args:
            package: 包名或pid进程号

        Returns:
            以字典返回所有内存信息
        """
        if not package:
            return None
        ret = {}
        _memory_info_count = 7
        if meminfo := self._get_app_meminfo(package):
            meminfo = self._parse_app_meminfo(meminfo)
            meminfo = meminfo.group('meminfo').strip()
            meminfo = meminfo.splitlines()
            # get meminfo head
            head_pattern = re.compile(r'(\S+)')
            headLine = [f'{t_1.lower()}_{t_2.lower()}' for t_1, t_2 in
                        zip(head_pattern.findall(meminfo[0].strip()), head_pattern.findall(meminfo[1].strip()))]

            pattern = re.compile(r'\s*(\S+\s?\S*)\s*(.*)\r?')
            for line in meminfo[3:]:
                line = pattern.search(line)
                if not line:
                    continue

                mem_pattern = re.compile(r'(\d+)\s*')
                mem = mem_pattern.findall(line.group(2).strip())
                # 将无内存信息的,补全0
                mem += [0 for _ in range(_memory_info_count - len(mem))]

                name = line.group(1).strip().lower().replace(' ', '_')

                ret[name] = {headLine[index]: v for index, v in enumerate([int(v) for v in mem])}
        return ret

    def get_app_summary(self, package: Union[str, int]) -> Optional[Dict[str, int]]:
        """
        获取app summary pss。
        返回一个包含java_heap/native_heap/code/stack/graphics/private_other/system/total/total_swap_pss的字典

        Args:
            package: 包名或pid进程号

        Returns:
            app内存信息概要
        """
        if not package:
            return None
        ret = {}
        if meminfo := self._get_app_meminfo(package):
            meminfo = self._parse_app_meminfo(meminfo)
            meminfo = meminfo.group('app_summary').strip()
            pattern = re.compile(r'\s*(\S+\s?\S*\s?\S*):\s*(\d+)')
            for v in pattern.findall(meminfo):
                name = v[0].strip().lower().replace(' ', '_')
                memory = int(v[1])
                ret[name] = memory
            return ret

    @staticmethod
    def _pares_memory(memory: str):
        """
        处理字符串memory,转换为int,单位KB

        Args:
            memory(str): 内存

        Returns:
            内存大小,单位(KB)
        """
        memory = int(memory.strip().replace(',', '').replace('K', ''))
        return memory

    @staticmethod
    def _parse_system_meminfo(meminfo: str) -> Optional[Match[str]]:
        """
        处理adb shell dumpsys meminfo返回的数据

        Args:
            meminfo: 内存数据

        Returns:
            包含uptime、realtime、pss_by_process、pss_by_ommAdjustemnt、pss_by_category、
            total_ram、free_ram、used_ram、lost_ram
        """
        pattern = re.compile(r'Uptime: (?P<uptime>\d+) Realtime: (?P<realtime>\d+)'
                             r'.*Total PSS by process:(?P<pss_by_process>.*)'
                             r'.*Total PSS by OOM adjustment:(?P<pss_by_ommAdjustemnt>.*)'
                             r'.*Total PSS by category:(?P<pss_by_category>.*)'
                             r'.*Total RAM:\s*(?P<total_ram>\S+)'
                             r'.*Free RAM:\s*(?P<free_ram>\S+)'
                             r'.*Used RAM:\s*(?P<used_ram>\S+)'
                             r'.*Lost RAM:\s*(?P<lost_ram>\S+)', re.DOTALL)
        return m if (m := pattern.search(meminfo)) else None

    @staticmethod
    def _parse_app_meminfo(meminfo: str):
        pattern = re.compile(r'Uptime: (?P<uptime>\d+) Realtime: (?P<realtime>\d+)'
                             r'.*\*\* MEMINFO in pid (?P<pid>\d+) \[(?P<package_name>\S+)] \*\*'
                             r'(?P<meminfo>.*)'
                             r'.*App Summary\s*(?P<app_summary>.*)'
                             r'.*Objects(.*)', re.DOTALL)
        return m if (m := pattern.search(meminfo)) else None

    def _get_app_meminfo(self, package: Union[str, int]):
        """
        'adb shell dumpsys meminfo <packageName|pid>' 获取指定包或进程号的内存信息

        Raises:
            AdbProcessNotFound: 未找到对应进程时弹出异常
        Args:
            package: 包名或pid进程号

        Returns:
            内存信息
        """
        if 'No process found for:' in (ret := self.device.shell(['dumpsys', 'meminfo', package])):
            raise AdbProcessNotFound(ret.strip())
        else:
            return ret

    def get_system_meminfo(self) -> str:
        """
        'adb shell dumpsys meminfo' 获取系统内存信息

        Returns:
            内存信息
        """
        return self.device.shell(['dumpsys', 'meminfo'])


if __name__ == '__main__':
    from adbutils import ADBDevice
    from adbutils.extra.performance.meminfo import Meminfo

    device = ADBDevice(device_id='')
    performance = Meminfo(device)

    performance.get_system_meminfo()

    for i in range(100):
        package_name = device.foreground_package
        total = performance.get_app_summary(package_name)['total']
        # 打印package_name对应app的内存占用
        print(f'{(int(total) / 1024):.0f}MB')



# -*- coding: utf-8 -*-
import re
import time
from typing import Union, Tuple, List, Optional, Dict

from adbutils import ADBDevice
from adbutils.constant import ANDROID_TMP_PATH, BUSYBOX_LOCAL_PATH, BUSYBOX_REMOTE_PATH
from adbutils._utils import get_std_encoding
from adbutils.extra.performance.exceptions import AdbNoInfoReturn

from loguru import logger


class Cpu(object):
    # user/nice/system/idle/iowait/irq/softirq/stealstolen/guest
    cpu_jiffies_pattern = re.compile(r'(\d+)')
    total_cpu_pattern = re.compile(r'cpu\s+(.*)')
    core_stat_pattern = re.compile(r'cpu(\d+)\s*(.*)')
    app_stat_pattern = re.compile(r'((\d+)\s*\(\S+\)(\s+\S+){50})+')

    def __init__(self, device: ADBDevice):
        self.device = device
        self._total_cpu_stat = []
        self._core_cpu_stat = []
        self._app_cpu_stat = {}

    def get_cpu_usage(self, name: Union[str, int, List[Union[int, str]], Tuple[Union[str, int], ...], None] = None):
        """
        获取设备cpu使用率

        Args:
            name: 根据name中的值查找对应pid的cpu使用率

        Returns:
            (total_cpu_usage, core_cpu_usage, app_cpu_usage)
            app_cpu_usage: 是一个以pid为索引的字典
            core_cpu_usage: 是一个列表,索引对应cpu核心
            total_cpu_usage: 一个float或int
        """
        # step1: 转换name为pid
        if isinstance(name, (str, int)):
            name = [name]
        pid_list = name and self._transform_name_to_pid(name) or []

        # step2: 运行命令,获得cpu信息
        try:
            total_cpu_stat, core_cpu_stat, app_cpu_stat = self._get_cpu_stat(pid_list)
        except AdbNoInfoReturn as err:
            logger.warning(err)
            return self.get_cpu_usage(name)

        _return_flag = 0
        if not self._total_cpu_stat or not self._core_cpu_stat:
            self._total_cpu_stat = total_cpu_stat
            self._core_cpu_stat = core_cpu_stat
            _return_flag = 1

        for _name in pid_list:
            if not self._app_cpu_stat.get(_name):
                if app_stat := app_cpu_stat.get(_name):
                    self._app_cpu_stat[_name] = app_stat
                    _return_flag = 2

        if _return_flag > 0:
            return self.get_cpu_usage(name)

        # step3: 计算总使用率
        total_idle = total_cpu_stat[3] - self._total_cpu_stat[3]
        total_cpu_time = sum(total_cpu_stat) - sum(self._total_cpu_stat)
        total_cpu_usage = 100 * (total_cpu_time - total_idle) / total_cpu_time

        # step4: 计算各核心使用率
        cpu_core_usage = []
        for cpu_index, core_stat in enumerate(core_cpu_stat):
            idle = core_stat[3] - self._core_cpu_stat[cpu_index][3]
            core_cpu_time = sum(core_stat) - sum(self._core_cpu_stat[cpu_index])
            cpu_core_usage.append(100 * (core_cpu_time - idle) / core_cpu_time)

        self._total_cpu_stat = total_cpu_stat
        self._core_cpu_stat = core_cpu_stat

        # step5: 计算各pid使用率
        app_usage_ret = {}

        for pid, app_stat in app_cpu_stat.items():
            if app_stat:
                app_cpu_time = sum([int(v) for v in app_stat[13:15]]) - \
                               sum([int(v) for v in self._app_cpu_stat[pid][13:15]])
                app_usage = 100 * (app_cpu_time / total_cpu_time)
                app_usage_ret[name[pid_list.index(pid)]] = app_usage
                self._app_cpu_stat[pid] = app_stat

        return total_cpu_usage, cpu_core_usage, app_usage_ret

    def _get_cpu_stat(self, name: Optional[List[int]] = None) -> \
            Tuple[List[int], List[List[int]], Dict[int, Optional[List[str]]]]:
        cmds = self._create_command(name)
        proc = self.device.start_shell(cmds)

        stdout, stderr = proc.communicate()

        stdout = stdout.decode(get_std_encoding(stdout))
        stderr = stderr.decode(get_std_encoding(stdout))

        app_cpu_stat = name and {pid: None for pid in name} or {}
        if ret := self.app_stat_pattern.findall(stdout):
            pattern = re.compile(r'(\S+)\s*')
            for v in ret:
                pid = int(v[1])
                app_cpu_stat[pid] = pattern.findall(v[0])
                stdout = stdout.replace(v[0], '')

        if not stdout:
            raise AdbNoInfoReturn(f'cpu信息获取异常')

        total_cpu_stat, core_cpu_stat = self._pares_cpu_stat(stdout)

        if not total_cpu_stat or not core_cpu_stat:
            return self._get_cpu_stat(name)

        return total_cpu_stat, core_cpu_stat, app_cpu_stat

    def _install_busyBox(self) -> None:
        """
        check if busybox installed

        Returns:
            None
        """
        if not self.device.check_file(ANDROID_TMP_PATH, 'busybox'):
            if 'v8' in self.device.abi_version:
                local = BUSYBOX_LOCAL_PATH.format('v8l')
            elif 'v7r' in self.device.abi_version:
                local = BUSYBOX_LOCAL_PATH.format('v7r')
            elif 'v7m' in self.device.abi_version:
                local = BUSYBOX_LOCAL_PATH.format('v7m')
            elif 'v7l' in self.device.abi_version:
                local = BUSYBOX_LOCAL_PATH.format('v7l')
            elif 'v5' in self.device.abi_version:
                local = BUSYBOX_LOCAL_PATH.format('v5l')
            else:
                local = BUSYBOX_LOCAL_PATH.format('v8l')

            self.device.push(local=local, remote=BUSYBOX_REMOTE_PATH)
            time.sleep(1)
            self.device.shell(['chmod', '755', BUSYBOX_REMOTE_PATH])

    def _pares_cpu_stat(self, stat: str) -> Tuple[List[int], List[List[int]]]:
        """
        处理cpu信息数据

        Args:
            stat: cpu数据

        Returns:
            总cpu数据和每个核心的数据
        """
        total_cpu_stat = None
        core_cpu_stat = []

        if total_stat := self.total_cpu_pattern.findall(stat):
            total_stat = self.cpu_jiffies_pattern.findall(total_stat[0])
            total_cpu_stat = [int(v) for v in total_stat]

        if core_stat_list := self.core_stat_pattern.findall(stat):
            for core_stat in core_stat_list:
                _core_stat = self.cpu_jiffies_pattern.findall(core_stat[1].strip())
                core_cpu_stat.append([int(v) for v in _core_stat])

        if not total_cpu_stat or not core_cpu_stat:
            raise AdbNoInfoReturn('cpu信息获取异常')

        return total_cpu_stat, core_cpu_stat

    @staticmethod
    def _create_command(name: Optional[List[int]] = None):
        """
        根据pid创建cmd命令

        Args:
            name: 包含pid的列表

        Returns:
            cmd命令
        """
        cmds = ['cat /proc/stat']
        if name:
            for pid in name:
                cmds += [f'cat /proc/{pid}/stat']

        return '&'.join(cmds)

    def _transform_name_to_pid(self, name: Union[List, Tuple]) -> Optional[List[int]]:
        """
        处理传参,将包名更改为pid

        Args:
            name: 需要处理的包名或pid

        Returns:
            只包含pid的列表
        """
        ret = []
        for _name in name:
            if isinstance(_name, str):
                if pid := self.device.get_pid_by_name(_name):
                    pid = pid[0][0]
                else:
                    logger.warning(f"应用:'{_name}'未运行")
                    pid = None
            elif isinstance(_name, int):
                pid = _name
            else:
                pid = None
            ret.append(pid)
        return ret


if __name__ == '__main__':
    from adbutils import ADBDevice
    from adbutils.extra.performance.cpu import Cpu

    device_id = ''
    device = ADBDevice(device_id=device_id)
    top_watcher = Cpu(device)

    while True:
        total_cpu_usage, cpu_core_usage, app_usage_ret = top_watcher.get_cpu_usage(device.foreground_package)
        print('cpu={} core={}, {}'.format(
            f'{total_cpu_usage:.1f}%',
            '\t'.join([f'cpu{core_index}:{usage:.1f}%' for core_index, usage in enumerate(cpu_core_usage)]),
            '\t'.join([f'{name}:{usage:.1f}%' for name, usage in app_usage_ret.items()])
        ))
        time.sleep(.5)



# -*- coding: utf-8 -*-
from adbutils.constant import ADB_INSTALL_FAILED


class AdbBaseError(Exception):
    def __init__(self, message):
        self.message = message

    def __repr__(self):
        return repr(self.message)


class AdbError(AdbBaseError):
    """ There was an exception that occurred while ADB command """
    def __init__(self, stdout, stderr, message: str = None):
        super(AdbError, self).__init__(message=message)
        self.stdout = stdout
        self.stderr = stderr

    def __repr__(self):
        return f"stdout[{self.stdout}] stderr[{self.stderr}]"


class AdbShellError(AdbError):
    """ There was an exception that occurred while ADB shell command """


class AdbSDKVersionError(AdbBaseError):
    """Errors caused by insufficient sdb versions """


class AdbTimeout(AdbBaseError):
    """ Adb command time out"""


class NoDeviceSpecifyError(AdbBaseError):
    """ No device was specified when ADB was commanded """


class AdbDeviceConnectError(AdbBaseError):
    """ Failed to connect device """
    CONNECT_ERROR = r"error:\s*(" \
                    r"(device \'\S+\' not found)|" \
                    r"(cannot connect to daemon at [\w\:\s\.]+ Connection timed out)|" \
                    r"(device offline))"


class AdbInstallError(AdbBaseError):
    """ An error while adb install apk failed """
    def __repr__(self):
        return repr(str(self))

    def __str__(self):
        if self.message in ADB_INSTALL_FAILED:
            return ADB_INSTALL_FAILED[self.message]
        else:
            return f'adb install failed,\n{self.message}'


class AdbExtraModuleNotFount(AdbBaseError):
    """ An error while adb extra-module not found"""



import random
import subprocess
import re
import socket
import os
import time
import warnings
#TODO: pm install/uninstall
import numpy as np
from baseImage import Rect, Point

from adbutils._utils import (get_adb_exe, split_cmd, _popen_kwargs, get_std_encoding, check_file)
from adbutils.constant import (ANDROID_ADB_SERVER_HOST, ANDROID_ADB_SERVER_PORT, ADB_CAP_RAW_REMOTE_PATH,
                               ADB_CAP_RAW_LOCAL_PATH, IP_PATTERN, ADB_DEFAULT_KEYBOARD, ANDROID_TMP_PATH,
                               ADB_KEYBOARD_APK_PATH)
from adbutils.exceptions import (AdbError, AdbShellError, AdbBaseError, AdbTimeout, NoDeviceSpecifyError,
                                 AdbDeviceConnectError, AdbInstallError, AdbSDKVersionError, AdbExtraModuleNotFount)
from adbutils._wraps import retries
from loguru import logger

from typing import Union, List, Optional, Tuple, Dict, Match, Iterator, Final


class ADBClient(object):
    SUBPROCESS_FLAG: Final[int] = _popen_kwargs()['creationflags']

    def __init__(self, device_id: Optional[str] = None, adb_path: Optional[str] = None,
                 host: Optional[str] = ANDROID_ADB_SERVER_HOST,
                 port: Optional[int] = ANDROID_ADB_SERVER_PORT):
        """
        Args:
            device_id (str): 指定设备名
            adb_path (str): 指定adb路径
            host (str): 指定连接地址
            port (int): 指定连接端口
        """
        self.device_id = device_id
        self.adb_path = adb_path or get_adb_exe()
        self._set_cmd_options(host, port)
        self.connect()

    @property
    def host(self) -> str:
        return self.__host

    @property
    def port(self) -> int:
        return self.__port

    def _set_cmd_options(self, host: str, port: int):
        """
        Args:
            host (str): 指定连接地址
            port (int): 指定连接端口
        """
        self.__host = host
        self.__port = port
        self.cmd_options = [self.adb_path]
        if self.host not in ('127.0.0.1', 'localhost'):
            self.cmd_options += ['-H', self.host]
        if self.port != ANDROID_ADB_SERVER_PORT:
            self.cmd_options += ['-P', str(self.port)]

    @property
    def server_version(self) -> int:
        """
        获得cmd version
        :return: adb server版本
        """
        ret = self.cmd('version', devices=False)
        pattern = re.compile(r'Android Debug Bridge version \d.\d.(\d+)')
        if version := pattern.findall(ret):
            return int(version[0])

    @property
    def devices(self) -> Dict[str, str]:
        """
        command 'adb devices'

        Returns:
            devices dict key[device_name]-value[device_state]
        """
        pattern = re.compile(r'([\S]+)\t([\w]+)\n?')
        ret = self.cmd("devices", devices=False)
        return {value[0]: value[1] for value in pattern.findall(ret)}

    def get_device_id(self, decode: bool = False) -> str:
        return decode and self.device_id.replace(':', '_') or self.device_id

    def start_server(self) -> None:
        """
        command 'adb start_server'

        Returns:
            None
        """
        self.cmd('start-server', devices=False)

    def kill_server(self) -> None:
        """
        command 'adb kill_server'

        Returns:
            None
        """
        self.cmd('kill-server', devices=False)

    @retries(2, exceptions=(AdbDeviceConnectError,))
    def connect(self, force: Optional[bool] = False) -> None:
        """
        command 'adb connect <device_id>'

        Args:
            force: 不判断设备当前状态,强制连接

        Returns:
                连接成功返回True,连接失败返回False
        """
        if self.device_id and ':' in self.device_id and (force or self.status != 'devices'):
            ret = self.cmd(f"connect {self.device_id}", devices=False, skip_error=True)
            if 'failed' in ret:
                raise AdbDeviceConnectError(f'failed to connect to {self.device_id}')

    def disconnect(self) -> None:
        """
        command 'adb -s <device_id> disconnect'

        Returns:
            None
        """
        if ':' in self.device_id:
            self.cmd(f"disconnect {self.device_id}", devices=False)

    def forward(self, local: str, remote: str, no_rebind: Optional[bool] = True) -> None:
        """
        command adb forward

        Args:
            local:  要转发的本地端口
            remote: 要与local绑定的设备端口
            no_rebind: if True,如果local端已经绑定则失败

        Returns:
            None
        """
        cmds = ['forward']
        if no_rebind:
            cmds += ['--no-rebind']
        self.cmd(cmds + [local, remote])

    def remove_forward(self, local: Optional[str] = None) -> None:
        """
        command adb forward --remove <local>

        Args:
            local: 本地端口。如果未指定local,则默认清除所有连接' adb forward --remove-all'

        Returns:
            None
        """
        if local:
            cmds = ['forward', '--remove', local]
        else:
            cmds = ['forward', '--remove-all']
        self.cmd(cmds)

    def get_forwards(self, device_id: Optional[str] = None) -> Dict[str, List[Tuple[str, str]]]:
        """
        command 'adb forward --list'

        Args:
            device_id (str): 获取指定设备下的端口
        Returns:
            forwards dict key[device_name]-value[Tuple[local, remote]]
        """
        forwards = {}
        pattern = re.compile(r'([\S]+)\s([\S]+)\s([\S]+)\n?')
        ret = self.cmd(['forward', '--list'], devices=False, skip_error=True)
        for value in pattern.findall(ret):
            if device_id and device_id != value[0]:
                continue
            if value[0] in forwards:
                forwards[value[0]] += [(value[1], value[2])]
            else:
                forwards[value[0]] = [(value[1], value[2])]

        return forwards

    def get_forward_port(self, remote: str, device_id: Optional[str] = None) -> Optional[int]:
        """
        获取开放端口的端口号

        Args:
            remote (str): 设备端口
            device_id (str): 获取指定设备下的端口
        Returns:
            本地端口号
        """
        forwards = self.get_forwards(device_id=device_id)
        local_pattern = re.compile(r'tcp:(\d+)')
        for device_id, value in forwards.items():
            if isinstance(value, (list, tuple)):
                for _local, _remote in value:
                    if (_remote == remote) and (ret := local_pattern.search(_local)):
                        return int(ret.group(1))
        return None

    def get_available_forward_local(self) -> int:
        """
        随机获取一个可用的端口号

        Returns:
            可用端口(int)
        """
        sock = socket.socket()
        port = random.randint(11111, 20000)
        result = False
        try:
            sock.bind((self.host, port))
            result = True
        except socket.error:
            pass
        sock.close()
        if result:
            return port
        return self.get_available_forward_local()

    def push(self, local: str, remote: str) -> None:
        """
        command 'adb push <local> <remote>'

        Args:
            local: 发送文件的路径
            remote: 发送到设备上的路径

        Raises:
            RuntimeError:文件不存在

        Returns:
            None
        """
        if not check_file(local):
            raise RuntimeError(f"file: {local} does not exists")
        self.cmd(['push', local, remote], decode=False)

    def pull(self, local: str, remote: str) -> None:
        """
        command 'adb pull <remote> <local>

        Args:
            local: 本地的路径
            remote: 设备上的路径

        Returns:
            None
        """
        self.cmd(['pull', remote, local], decode=False)

    def install(self, local: str, install_options: Union[str, list, None] = None) -> bool:
        """
        command 'adb install <local>'

        Args:
            local: apk文件路径
            install_options: 可指定参数
                    "-r",  # 重新安装现有应用，并保留其数据。
                    "-t",  # 允许安装测试 APK。
                    "-g",  # 授予应用清单中列出的所有权限。
                    "-d",  # 允许APK降级覆盖安装
                    "-l",  # 将应用安装到保护目录/mnt/asec
                    "-s",  # 将应用安装到sdcard
        Raises:
            AdbInstallError: 安装失败
            AdbError: 安装失败
        Returns:
            安装成功返回True
        """
        cmds = ['install']
        if isinstance(install_options, str):
            cmds.append(install_options)
        elif isinstance(install_options, list):
            cmds = cmds + install_options

        cmds = cmds + [local]
        proc = self.start_cmd(cmds)
        stdout, stderr = proc.communicate()

        stdout = stdout.decode(get_std_encoding(stdout))
        stderr = stderr.decode(get_std_encoding(stdout))

        pattern = re.compile(r"Failure \[(.+):.+\]")
        if proc.returncode == 0:
            return True
        elif pattern.search(stderr):
            raise AdbInstallError(pattern.findall(stderr)[0])
        else:
            raise AdbError(stdout, stderr)

    def uninstall(self, package_name: str, install_options: Optional[str] = None) -> None:
        """
        command 'adb uninstall <package>

        Args:
            package_name: 需要卸载的包名
            install_options: 可指定参数
                            "-k",  # 移除软件包后保留数据和缓存目录。

        Returns:
            None
        """
        cmds = ['uninstall']
        if install_options and isinstance(install_options, str):
            cmds.append(install_options)

        cmds = cmds + [package_name]
        self.cmd(cmds)

    @property
    def status(self) -> Optional[str]:
        """
        command adb -s <device_id> get-state,返回当前设备状态

        Returns:
            当前设备状态
        """
        proc = self.start_cmd('get-state')
        stdout, stderr = proc.communicate()

        stdout = stdout.decode(get_std_encoding(stdout))
        stderr = stderr.decode(get_std_encoding(stdout))

        if proc.returncode == 0:
            return stdout.strip()
        elif "not found" in stderr:
            return None
        elif 'device offline' in stderr:
            return 'offline'
        else:
            raise AdbError(stdout, stderr)

    def cmd(self, cmds: Union[list, str], devices: Optional[bool] = True, decode: Optional[bool] = True,
            timeout: Optional[int] = None, skip_error: Optional[bool] = False):
        """
        创建cmd命令, 并返回命令返回值

        Args:
            cmds (list,str): 需要运行的参数
            devices (bool): 如果为True,则需要指定device-id,命令中会传入-s
            decode (bool): 是否解码stdout,stderr
            timeout (int): 设置命令超时时间
            skip_error (bool): 是否跳过报错
        Raises:
            AdbDeviceConnectError: 设备连接异常
            AdbTimeout:输入命令超时
        Returns:
            返回命令结果stdout
        """

        proc = self.start_cmd(cmds, devices)
        if timeout and isinstance(timeout, int):
            try:
                stdout, stderr = proc.communicate(timeout=timeout)
            except subprocess.TimeoutExpired:
                proc.kill()
                _, stderr = proc.communicate()
                raise AdbTimeout(f"cmd command {' '.join(proc.args)} time out")
        else:
            stdout, stderr = proc.communicate()

        if decode:
            stdout = stdout.decode(get_std_encoding(stdout))
            stderr = stderr.decode(get_std_encoding(stderr))

        if proc.returncode > 0:
            pattern = AdbDeviceConnectError.CONNECT_ERROR
            if isinstance(stderr, bytes):
                pattern = pattern.encode("utf-8")
            if re.search(pattern, stderr):
                raise AdbDeviceConnectError(stderr)
            if not skip_error:
                raise AdbError(stdout, stderr)

        return stdout

    def start_cmd(self, cmds: Union[list, str], devices: bool = True) -> subprocess.Popen:
        """
        根据cmds创建一个Popen

        Args:
            cmds: cmd commands
            devices: 如果为True,则需要指定device-id,命令中会传入-s
        Raises:
            NoDeviceSpecifyError:没有指定设备运行cmd命令
        Returns:
            Popen管道
        """
        cmds = split_cmd(cmds)
        if devices:
            if not self.device_id:
                raise NoDeviceSpecifyError('must set device_id')
            cmd_options = self.cmd_options + ['-s', self.device_id]
        else:
            cmd_options = self.cmd_options

        cmds = cmd_options + cmds
        logger.info(' '.join(cmds))
        proc = subprocess.Popen(
            cmds,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            creationflags=self.SUBPROCESS_FLAG
        )
        return proc


class ADBShell(ADBClient):
    SHELL_ENCODING: Final[str] = 'utf-8'  # adb shell的编码
    PS_HEAD: Final[List[str]] = ['user', 'pid', 'ppid', 'vsize', 'rss', '', 'wchan', 'pc', 'name']  # adb shell ps

    @property
    def line_breaker(self) -> str:
        """
        Set carriage return and line break property for various platforms and SDK versions

        Returns:
            carriage return and line break string
        """
        if not hasattr(self, '_line_breaker'):
            if self.sdk_version >= 24:
                line_breaker = os.linesep
            else:
                line_breaker = '\r' + os.linesep
            line_breaker = line_breaker.encode("ascii")
            setattr(self, '_line_breaker', line_breaker)

        return getattr(self, '_line_breaker')

    @property
    def memory(self) -> str:
        """
        获取设备内存大小

        Returns:
            单位MB
        """
        ret = self.shell(['dumpsys', 'meminfo'])
        pattern = re.compile(r'.*Total RAM:\s+(\S+)\s+', re.DOTALL)
        if m := pattern.search(ret):
            memory = m.group(1)
        else:
            raise AdbBaseError(ret)

        if ',' in memory:
            memory = memory.split(',')
            # GB: memory[0], MB: memory[1], KB: memory[2]
            memory = int(int(memory[0]) * 1024) + int(memory[1])
        else:
            memory = round(int(memory) / 1024)
        return f'{str(memory)}MB'

    @property
    def cpu_coreNum(self) -> Optional[int]:
        """
        获取cpu核心数量

        Returns:
            cpu核心数量
        """
        if not hasattr(self, '_cpu_coreNum'):
            setattr(self, '_cpu_coreNum', int(self.shell("cat /proc/cpuinfo").strip().count('processor')))

        return getattr(self, '_cpu_coreNum')

    @property
    def cpu_max_freq(self) -> List[Optional[int]]:
        """
        获取cpu各核心的最高频率

        Raises:
            AdbBaseError: 获取cpu信息失败
        Returns:
            包含核心最高频率的列表
        """
        _cores = []
        cmds = [f"cat /sys/devices/system/cpu/cpu{i}/cpufreq/scaling_max_freq" for i in range(self.cpu_coreNum)]
        cmds = '&'.join(cmds)

        ret = self.shell(cmds)
        if not ret:
            raise AdbBaseError('get cpufreq error')

        pattern = re.compile('(\d+)')
        if ret := pattern.findall(ret):
            _cores = [int(int(core) / 1000) for core in ret]
        else:
            raise AdbBaseError('get cpufreq error')

        return _cores

    @property
    def cpu_min_freq(self) -> List[Optional[int]]:
        """
        获取cpu各核心的最低频率

        Raises:
            AdbBaseError: 获取cpu信息失败
        Returns:
            包含核心最低频率的列表
        """
        _cores = []
        cmds = [f"cat /sys/devices/system/cpu/cpu{i}/cpufreq/scaling_min_freq" for i in range(self.cpu_coreNum)]
        cmds = '&'.join(cmds)

        ret = self.shell(cmds)
        if not ret:
            raise AdbBaseError('get cpufreq error')

        pattern = re.compile('(\d+)')
        if ret := pattern.findall(ret):
            _cores = [int(int(core) / 1000) for core in ret]
        else:
            raise AdbBaseError('get cpufreq error')

        return _cores

    @property
    def cpu_cur_freq(self) -> List[Optional[int]]:
        """
        获取cpu各核心的当前频率

        Raises:
            AdbBaseError: 获取cpu信息失败
        Returns:
            包含核心当前频率的列表
        """
        _cores = []
        cmds = [f"cat /sys/devices/system/cpu/cpu{i}/cpufreq/scaling_cur_freq" for i in range(self.cpu_coreNum)]
        cmds = '&'.join(cmds)

        ret = self.shell(cmds)
        if not ret:
            raise AdbBaseError('get cpufreq error')

        pattern = re.compile('(\d+)')
        if ret := pattern.findall(ret):
            _cores = [int(int(core) / 1000) for core in ret]
        else:
            raise AdbBaseError('get cpufreq error')

        return _cores

    @property
    def cpu_abi(self) -> str:
        """
        获取cpu构架

        Returns:
            cpu构建
        """
        if not hasattr(self, '_cpu_adi'):
            setattr(self, '_cpu_adi', self.shell("getprop ro.product.cpu.abi").strip())

        return getattr(self, '_cpu_adi')

    @property
    def gpu_model(self):
        """
        获取gpu型号

        Returns:
            gpu型号
        """
        if not hasattr(self, '_gpu_model'):
            ret = self.shell('dumpsys SurfaceFlinger')
            pattern = re.compile(r'GLES:\s+(.*)')
            m = pattern.search(ret)
            if not m:
                return None
            _list = m.group(1).split(',')
            gpuModel = ''

            if len(_list) > 0:
                gpuModel = _list[1].strip()
            setattr(self, '_gpu_model', gpuModel)

        return getattr(self, '_gpu_model')

    @property
    def opengl_version(self):
        """
        获取设备opengl版本

        Returns:
            opengl版本
        """
        if not hasattr(self, '_opengl'):
            ret = self.shell('dumpsys SurfaceFlinger')
            pattern = re.compile(r'GLES:\s+(.*)')
            m = pattern.search(ret)
            if not m:
                return None
            _list = m.group(1).split(',')
            opengl = ''

            if len(_list) > 1:
                m2 = re.search(r'(\S+\s+\S+\s+\S+).*', _list[2])
                if m2:
                    opengl = m2.group(1)
            setattr(self, '_opengl', opengl)

        return getattr(self, '_opengl')

    @property
    def model(self) -> str:
        """
        获取手机型号

        Returns:
            手机型号
        """
        if not hasattr(self, '_model'):
            setattr(self, '_model', self.getprop('ro.product.model'))

        return getattr(self, '_model')

    @property
    def manufacturer(self) -> str:
        """
        获取手机厂商名

        Returns:
            手机厂商名
        """
        if not hasattr(self, '_manufacturer'):
            setattr(self, '_manufacturer', self.getprop('ro.product.manufacturer'))

        return getattr(self, '_manufacturer')

    @property
    def android_version(self) -> str:
        """
        获取系统安卓版本

        Returns:
            安卓版本
        """
        if not hasattr(self, '_android_version'):
            setattr(self, '_android_version', self.getprop('ro.build.version.release'))

        return getattr(self, '_android_version')

    @property
    def sdk_version(self) -> int:
        """
        获取sdk版本

        Returns:
            sdk版本号
        """
        if not hasattr(self, '_sdk_version'):
            setattr(self, '_sdk_version', int(self.getprop('ro.build.version.sdk')))

        return getattr(self, '_sdk_version')

    @property
    def abi_version(self) -> str:
        """
        获取abi版本

        Returns:
            abi版本
        """
        if not hasattr(self, '_abi_version'):
            setattr(self, '_abi_version', self.getprop('ro.product.cpu.abi'))

        return getattr(self, '_abi_version')

    @property
    def displayInfo(self) -> Dict[str, Union[int, float]]:
        """
        获取屏幕数据

        Returns:
            width/height/density/orientation/rotation/max_x/max_y
        """
        display_info = self.getPhysicalDisplayInfo()
        orientation = self.orientation
        max_x, max_y = self.getMaxXY()
        display_info.update({
            "orientation": orientation,
            "rotation": orientation * 90,
            "max_x": max_x,
            "max_y": max_y,
        })
        return display_info

    @property
    def dpi(self) -> int:
        if ret := self.getprop('ro.sf.lcd_density', True):
            return int(ret)

    @property
    def orientation(self) -> int:
        """
        获取屏幕方向

        Returns:
            屏幕方向 0/1/2
        """
        return self.getDisplayOrientation()

    @property
    def ip_address(self) -> Optional[str]:
        """
        获得设备ip地址

        Returns:
            未找到则返回None,找到则返回IP address
        """

        def get_ip_address_from_interface(interface):
            # android >= 6.0: ip -f inet addr show {interface}
            try:
                res = self.shell(f'ip -f inet addr show {interface}')
            except AdbShellError:
                res = ''
            if matcher := re.search(r"inet (?P<ip>(\d+\.){3}\d+)", res):
                return matcher.group('ip')

            # android >= 6.0 backup method: ifconfig
            try:
                res = self.shell('ifconfig')
            except AdbShellError:
                res = ''
            if matcher := re.search(interface + r'.*?inet addr:((\d+\.){3}\d+)', res, re.DOTALL):
                return matcher.group(1)

            # android <= 6.0: netcfg
            try:
                res = self.shell('netcfg')
            except AdbShellError:
                res = ''
            if matcher := re.search(interface + r'.* ((\d+\.){3}\d+)/\d+', res):
                return matcher.group(1)

            # android <= 6.0 backup method: getprop dhcp.{}.ipaddress
            try:
                res = self.shell('getprop dhcp.{}.ipaddress'.format(interface))
            except AdbShellError:
                res = ''
            if matcher := IP_PATTERN.search(res):
                return matcher.group(0)

            # sorry, no more methods...
            return None

        interfaces = ('eth0', 'eth1', 'wlan0')
        for i in interfaces:
            ip = get_ip_address_from_interface(i)
            if ip and not ip.startswith('172.') and not ip.startswith('127.') and not ip.startswith('169.'):
                return ip
        return None

    @property
    def foreground_activity(self) -> str:
        """
        获取当前前台activity

        Raises:
            AdbBaseError: 没有获取到前台activity
        Returns:
            当前activity
        """
        if m := self._get_activityRecord(key='mResumedActivity'):
            return m.group('activity')
        else:
            raise AdbBaseError(f'running_activities get None')

    @property
    def foreground_package(self) -> str:
        """
        获取当前前台包名

        Raises:
            AdbBaseError: 没有获取到前台包名
        Returns:
            当前包名
        """
        if m := self._get_activityRecord(key='mResumedActivity'):
            return m.group('packageName')
        else:
            raise AdbBaseError(f'running_activities get None')

    @property
    def running_activities(self) -> List[str]:
        """
        获取正在运行的所有activity

        Raises:
            AdbBaseError: 未获取到当前运行的activity
        Returns:
            所有正在运行的activity
        """
        if m := self._get_running_activities():
            return [match.group('activity') for match in m]
        else:
            raise AdbBaseError(f'running_activities get None')

    @property
    def running_package(self) -> List[str]:
        """
        获取正在运行的所有包名

        Raises:
            AdbBaseError: 未获取到当前运行的包名
        Returns:
            所有正在运行的包名
        """
        if m := self._get_running_activities():
            return [match.group('packageName') for match in m]
        else:
            raise AdbBaseError(f'running_activities get None')

    @property
    def default_ime(self) -> str:
        """
        获取当前输入法ID

        Returns:
            输入法ID
        """
        try:
            ime = self.shell(['settings', 'get', 'secure', 'default_input_method']).strip()
        except AdbShellError:
            ime = None
        return ime

    @property
    def ime_list(self) -> List[str]:
        """
        获取系统可用输入法

        Returns:
            输入法列表
        """
        if ret := self.shell(['ime', 'list', '-s']):
            return ret.split()

    def is_keyboard_shown(self) -> bool:
        """
        判断键盘是否显示

        Returns:
            True显示键盘/False未显示键盘
        """

        if ret := self.shell(['dumpsys', 'input_method']):
            return 'mInputShown=true' in ret
        return False

    def is_screenon(self) -> bool:
        """
        检测屏幕打开关闭状态

        Raises:
            AdbBaseError: 未检测到屏幕打开状态
        Returns:
            True屏幕打开/False屏幕关闭
        """
        pattern = re.compile(r'mScreenOnFully=(?P<Bool>true|false)')
        ret = self.shell(['dumpsys', 'window', 'policy'])

        if m := pattern.search(ret):
            return m.group('Bool') == 'true'
        else:
            # MIUI11
            screenOnRE = re.compile('screenState=(SCREEN_STATE_ON|SCREEN_STATE_OFF)')
            m = screenOnRE.search(self.shell('dumpsys window policy'))
            if m:
                return m.group(1) == 'SCREEN_STATE_ON'
        raise AdbBaseError('Could not determine screen ON state')

    def is_locked(self) -> bool:
        """
        判断屏幕是否锁定

        Raises:
            AdbBaseError: 未检测到屏幕锁定状态
        Returns:
            True屏幕锁定/False屏幕未锁定
        """
        ret = self.shell('dumpsys window policy')
        pattern = re.compile(r'(?:mShowingLockscreen|isStatusBarKeyguard|showing)=(?P<Bool>true|false)')

        if m := pattern.search(ret):
            return m.group('Bool') == 'true'
        raise AdbBaseError('Could not determine screen lock state')

    def get_pid_by_name(self, packageName: str, fuzzy_search: Optional[bool] = False) -> List[Tuple[int, str]]:
        """
        根据进程名获取对应pid

        Args:
            packageName: 包名
            fuzzy_search: if True,返回所有包含packageName的pid
        Returns:
            获取到的进程列表(pid, name)
        """
        ps_len = len(self.PS_HEAD)
        if fuzzy_search:
            return [(int(proc[self.PS_HEAD.index('pid')]), proc_name) for proc in self.get_process()
                    if len(proc) == ps_len and (packageName in (proc_name := proc[self.PS_HEAD.index('name')]))]
        else:
            return [(int(proc[self.PS_HEAD.index('pid')]), proc_name) for proc in self.get_process()
                    if len(proc) == ps_len and (packageName == (proc_name := proc[self.PS_HEAD.index('name')]))]

    def get_process(self, flag_options: Union[str, list, tuple, None] = None) -> List[List[str]]:
        """
        command 'adb shell ps'

        Returns:
            所有进程的列表
        """
        cmds = ['ps']
        if isinstance(flag_options, str):
            cmds.append(flag_options)
        elif isinstance(flag_options, (list, tuple)):
            cmds = cmds + flag_options

        process = []
        process_pattern = re.compile('(\S+)')
        if ret := self.shell(cmds):
            ret = ret.splitlines()
            for v in ret[1:]:
                if proc := process_pattern.findall(v):
                    process.append(proc)

        return process

    def _get_running_activities(self) -> Optional[List[Match[str]]]:
        """
        command 'adb dumpsys activity activities'
        获取各个Stack中正在运行的activities参数

        Returns:
            包含了多个Match的列表, Match可以使用memory/user/packageName/activity/task
        """
        running_activities = []
        cmds = ['dumpsys', 'activity', 'activities']
        activities = self.shell(cmds)
        # 获取Stack
        pattern = re.compile(r'Stack #([\d]+):')
        stack = pattern.findall(activities)
        if not stack:
            return None
        stack.sort()
        # 根据Stack拆分running activities
        for index in stack:
            pattern = re.compile(rf'Stack #{index}[\s\S]+?Running activities \(most recent first\):([\s\S]+?)\r\n\r\n')
            ret = pattern.findall(activities)
            if ret:
                running_activities.append(ret[0])

        ret = []
        pattern = re.compile(
            r"TaskRecord[\s\S]+?Run #(?P<index>\d+):[\s]?"
            r"ActivityRecord\{(?P<memory>.*) (?P<user>.*) (?P<packageName>.*)/(?P<activity>\.?.*) (?P<task>.*)}")
        for v in running_activities:
            if m := pattern.search(v):
                ret.append(m)

        return ret

    def _get_activityRecord(self, key: str) -> Optional[Match[str]]:
        """
        command 'adb dumpsys activity activities'
        根据<key>获取对应ActivityRecord信息

        Returns:
            Match,可以使用memory/user/packageName/activity/task
        """
        cmds = ['dumpsys', 'activity', 'activities']
        ret = self.shell(cmds)
        pattern = re.compile(
            rf'{key}: '
            r'ActivityRecord\{(?P<memory>.*) (?P<user>.*) (?P<packageName>.*)/\.?(?P<activity>.*) (?P<task>.*)}[\n\r]')

        if m := pattern.search(ret):
            return m
        else:
            return None

    def check_app(self, name: str) -> bool:
        """
        判断应用是否安装

        Args:
            name: package name

        Returns:
            return True if find, false otherwise
        """
        if name in self.app_list():
            return True
        return False

    def check_file(self, path: str, name: str) -> bool:
        """
        command 'adb shell find <name> in the <path>'

        Args:
            path: 在设备上的路径
            name: 需要检索的文件名

        Returns:
            bool 是否找到文件
        """
        return bool(self.raw_shell(['find', path, '-name', name]))

    def getMaxXY(self) -> Tuple[int, int]:
        """
        获取屏幕可点击的最大长宽距离

        Returns:
            max_x,max_y
        """
        ret = self.shell(['getevent', '-p']).split('\n')
        max_x, max_y = None, None
        pattern = re.compile(r'max ([0-9]+)')
        for i in ret:
            if i.find('0035') != -1:
                if ret := pattern.findall(i):
                    max_x = int(ret[0])

            if i.find('0036') != -1:
                if ret := pattern.findall(i):
                    max_y = int(ret[0])
        return max_x, max_y

    def getPhysicalDisplayInfo(self) -> Dict[str, Union[int, float]]:
        """
        Get value for display dimension and density from `mPhysicalDisplayInfo` value obtained from `dumpsys` command.

        Returns:
            physical display info for dimension and density

        """
        phyDispRE = re.compile(
            r'.*PhysicalDisplayInfo{(?P<width>\d+) x (?P<height>\d+), .*, density (?P<density>[\d.]+).*')
        ret = self.raw_shell('dumpsys display')
        if m := phyDispRE.search(ret):
            displayInfo = {}
            for prop in ['width', 'height']:
                displayInfo[prop] = int(m.group(prop))
            for prop in ['density']:
                # In mPhysicalDisplayInfo density is already a factor, no need to calculate
                displayInfo[prop] = float(m.group(prop))
            return displayInfo

        # This could also be mSystem or mOverscanScreen
        phyDispRE = re.compile('\s*mUnrestrictedScreen=\((?P<x>\d+),(?P<y>\d+)\) (?P<width>\d+)x(?P<height>\d+)')
        # This is known to work on older versions (i.e. API 10) where mrestrictedScreen is not available
        dispWHRE = re.compile(r'\s*DisplayWidth=(?P<width>\d+) *DisplayHeight=(?P<height>\d+)')
        ret = self.raw_shell('dumpsys window')
        m = phyDispRE.search(ret, 0)
        if not m:
            m = dispWHRE.search(ret, 0)
        if m:
            displayInfo = {}
            for prop in ['width', 'height']:
                displayInfo[prop] = int(m.group(prop))
            for prop in ['density']:
                d = self._getDisplayDensity(strip=True)
                if d:
                    displayInfo[prop] = d
                else:
                    # No available density information
                    displayInfo[prop] = -1.0
            return displayInfo

        # gets C{mPhysicalDisplayInfo} values from dumpsys. This is a method to obtain display dimensions and density
        phyDispRE = re.compile(r'Physical size: (?P<width>\d+)x(?P<height>\d+).*Physical density: (?P<density>\d+)',
                               re.S)
        ret = self.raw_shell('wm size; wm density')

        if m := phyDispRE.search(ret):
            displayInfo = {}
            for prop in ['width', 'height']:
                displayInfo[prop] = int(m.group(prop))
            for prop in ['density']:
                displayInfo[prop] = float(m.group(prop))
            return displayInfo

        return {}

    def _getDisplayDensity(self, strip=True) -> Union[float, int]:
        """
        Get display density

        Args:
            strip: strip the output
        Returns:
            display density
        """
        BASE_DPI = 160.0

        if density := self.getprop('ro.sf.lcd_density', strip):
            return float(density) / BASE_DPI

        if density := self.getprop('qemu.sf.lcd_density', strip):
            return float(density) / BASE_DPI
        return -1.0

    def getDisplayOrientation(self) -> int:
        """
        Another way to get the display orientation, this works well for older devices (SDK version 15)

        Returns:
            display orientation information

        """
        # another way to get orientation, for old sumsung device(sdk version 15)
        SurfaceFlingerRE = re.compile(r'orientation=(\d+)')
        ret = self.shell('dumpsys SurfaceFlinger')
        if m := SurfaceFlingerRE.search(ret):
            return int(m.group(1))

        # Fallback method to obtain the orientation
        # See https://github.com/dtmilano/AndroidViewClient/issues/128
        surfaceOrientationRE = re.compile(r'SurfaceOrientation:\s+(\d+)')
        ret = self.shell('dumpsys input')
        if m := surfaceOrientationRE.search(ret):
            return int(m.group(1))
        # We couldn't obtain the orientation
        warnings.warn("Could not obtain the orientation, return 0")
        return 0

    def keyevent(self, keycode: Union[str, int]) -> None:
        """
        command 'adb shell input keyevent'
        Args:
            keycode: key code number or name

        Returns:
            None
        """
        self.shell(['input', 'keyevent', str(keycode)])

    def getprop(self, key: str, strip: Optional[bool] = True) -> Optional[str]:
        """
        command 'adb shell getprop <key>

        Args:
            key: 需要查询的参数
            strip: 删除文本头尾空格

        Returns:
            getprop获取到的参数
        """
        ret = self.raw_shell(['getprop', key])
        return strip and ret.rstrip() or ret

    def get_app_install_path(self, packageName: str) -> Optional[str]:
        """
        command 'adb shell pm path <package>'

        Args:
            packageName: 需要查找的包名

        Returns:
            包安装路径
        """
        if packageName in self.app_list():
            stdout = self.shell(['pm', 'path', packageName])
            if 'package:' in stdout:
                return stdout.split('package:')[1].strip()
        else:
            return None

    def app_list(self, flag_options: Union[str, list, tuple, None] = None) -> List[str]:
        """
        command 'adb shell pm list packages'

        Args:
            flag_options: 可指定参数
                    "-f",  # 查看它们的关联文件。
                    "-d",  # 进行过滤以仅显示已停用的软件包。
                    "-e",  # 进行过滤以仅显示已启用的软件包。
                    "-s",  # 进行过滤以仅显示系统软件包。
                    "-3",  # 进行过滤以仅显示第三方软件包。
                    "-i",  # 查看软件包的安装程序。
                    "-u",  # 也包括已卸载的软件包。
                    "--user user_id",  # 要查询的用户空间。

        Returns:

        """
        if self.sdk_version >= 24:
            cmds = ['cmd', 'package', 'list', 'packages']
        else:
            cmds = ['pm', 'list', 'packages']

        if isinstance(flag_options, str):
            cmds.append(flag_options)
        elif isinstance(flag_options, (list, tuple)):
            cmds = cmds + flag_options
        ret = self.shell(cmds)
        packages = ret.splitlines()
        # remove all empty string; "package:xxx" -> "xxx"
        packages = [p.split(":")[1] for p in packages if p]
        return packages

    def broadcast(self, action: str, user: str = None) -> None:
        """
        发送广播信号

        Args:
            action: 需要触发的广播行为
            user: 向指定组件广播

        Returns:
            None
        """
        cmds = ['am', 'broadcast'] + ['-a', action]
        if user:
            cmds += ['-n', user]
        self.start_cmd(cmds)

    def shell(self, cmds: Union[list, str], decode: Optional[bool] = True, skip_error: Optional[bool] = False) \
            -> Union[str, bytes]:
        """
        command 'adb shell

        Args:
            cmds (list,str): 需要运行的参数
            decode (bool): 是否解码stdout,stderr
            skip_error (bool): 是否跳过报错
        Raises:
            AdbShellError:指定shell命令时出错
        Returns:
            命令返回结果
        """
        if self.sdk_version < 25:
            # sdk_version < 25, adb shell 不返回错误
            # https://issuetracker.google.com/issues/36908392
            cmds = split_cmd(cmds) + [';', 'echo', '---$?---']
            ret = self.raw_shell(cmds, decode=decode).rstrip()
            if m := re.match("(.*)---(\d+)---$", ret, re.DOTALL):
                stdout = m.group(1)
                returncode = int(m.group(2))
            else:
                warnings.warn('return code not matched')
                stdout = ret
                returncode = 0

            if returncode > 0:
                if not skip_error:
                    raise AdbShellError(stdout, stderr=None)
            return stdout
        else:
            try:
                ret = self.raw_shell(cmds, decode=decode, skip_error=skip_error)
            except AdbError as err:
                raise AdbShellError(err.stdout, err.stderr)
            else:
                return ret

    def raw_shell(self, cmds: Union[list, str], decode: Optional[bool] = True, skip_error: Optional[bool] = False) \
            -> str:
        """
        command 'adb shell

        Args:
            cmds (list): 需要运行的参数
            decode (bool): 是否解码stdout,stderr
            skip_error (bool): 是否跳过报错
        Returns:
            命令返回结果
        """
        cmds = ['shell'] + split_cmd(cmds)
        stdout = self.cmd(cmds, decode=False, skip_error=skip_error)
        if not decode:
            return stdout

        try:
            return stdout.decode(self.SHELL_ENCODING)
        except UnicodeDecodeError:
            return str(repr(stdout))

    def start_shell(self, cmds: Union[list, str]):
        cmds = ['shell'] + split_cmd(cmds)
        return self.start_cmd(cmds)


class ADBDevice(ADBShell):
    def __init__(self, device_id: Optional[str] = None, adb_path: Optional[str] = None,
                 host: Optional[str] = ANDROID_ADB_SERVER_HOST,
                 port: Optional[int] = ANDROID_ADB_SERVER_PORT):
        """
        Args:
            device_id (str): 指定设备名
            adb_path (str): 指定adb路径
            host (str): 指定连接地址
            port (int): 指定连接端口
        """
        super(ADBDevice, self).__init__(device_id=device_id, adb_path=adb_path, host=host, port=port)
        self.set_input_method(ime_method=ADB_DEFAULT_KEYBOARD, ime_apk_path=ADB_KEYBOARD_APK_PATH)

    def screenshot(self, rect: Union[Rect, Tuple[int, int, int, int], List[int]] = None) -> np.ndarray:
        """
        command 'adb screencap'

        Args:
            rect: 自定义截取范围 Rect/(x, y, width, height)

        Raises:
                ValueError:传入参数rect错误
                OverflowError:rect超出屏幕边界范围
        Returns:
            图像数据
        """
        remote_path = ADB_CAP_RAW_REMOTE_PATH
        raw_local_path = ADB_CAP_RAW_LOCAL_PATH.format(device_id=self.get_device_id(True))

        self.raw_shell(['screencap', remote_path])
        self.start_shell(['chmod', '755', remote_path])
        self.pull(local=raw_local_path, remote=remote_path)

        # read size
        img_data = np.fromfile(raw_local_path, dtype=np.uint16)
        width, height = img_data[2], img_data[0]
        # read raw
        _line = 4  # 色彩通道数
        img_data = np.fromfile(raw_local_path, dtype=np.uint8)
        img_data = img_data[slice(_line * 4, len(img_data))]
        # 范围截取
        img_data = img_data.reshape(width, height, _line)
        width, height = img_data.shape[1::-1]
        if rect:
            if isinstance(rect, Rect):
                pass
            elif isinstance(rect, (tuple, list)):
                try:
                    rect = Rect(*rect)
                except TypeError:
                    raise ValueError('param "rect" takes 4 positional arguments <x,y,width,height>')
            else:
                raise ValueError('param "rect" must be <Rect>/tuple/list')

            # 判断边界是否超出width,height
            if not Rect(0, 0, width, height).contains(rect):
                raise OverflowError(f'rect不能超出屏幕边界 {rect}')
            x_min, y_min = int(rect.tl.x), int(rect.tl.y)
            x_max, y_max = int(rect.br.x), int(rect.br.y)
            img_data = img_data[y_min:y_max, x_min:x_max]

        img_data = img_data[:, :, ::-1][:, :, 1:4]  # imgData中rgbA转为ABGR,并截取bgr
        # 删除raw临时文件
        os.remove(raw_local_path)
        return img_data

    def start_app(self, package: str, activity: Optional[str] = None):
        """
        if not activity command 'adb shell monkey'
        if activity command 'adb shell am start

        Args:
            package: package name
            activity: activity name

        Returns:
            None
        """
        if not activity:
            cmds = ['monkey', '-p', package, '-c', 'android.intent.category.LAUNCHER', '1']
        else:
            cmds = ['am', 'start', '-n', f'{package}/{package}.{activity}']
        self.shell(cmds)

    def stop_app(self, package: str) -> None:
        """
        command 'adb shell am force-stop' to force stop the application

        Args:
            package: package name

        Returns:
            None
        """
        self.shell(['am', 'force-stop', package])

    def clear_app(self, package: str) -> None:
        """
        command 'adb shell pm clear' to force stop the application
        这条命令的效果相当于在设置里的应用信息界面点击了「清除缓存」和「清除数据」

        Args:
            package: package name

        Returns:
            None
        """
        self.shell(['pm', 'clear', package])

    def tap(self, point: Union[Tuple[int, int], Point]):
        """
        command 'adb shell input tap' 点击屏幕

        Args:
            point: 坐标(x,y)

        Returns:
            None
        """
        x, y = None, None
        if isinstance(point, Point):
            x, y = point.x, point.y
        elif isinstance(point, (tuple, list)):
            x, y = point[0], point[1]
        self.shell(f'input tap {x} {y}')

    def swipe(self, start_point: Union[Tuple[int, int], Point], end_point: Union[Tuple[int, int], Point],
              duration: int = 500) -> None:
        """
        command 'adb shell input swipe> 滑动屏幕

        Args:
            start_point: 起点坐标
            end_point: 重点坐标
            duration: 操作后延迟
        Returns:
            None
        """

        def _handle(point):
            if isinstance(point, Point):
                return point.x, point.y
            elif isinstance(point, (tuple, list)):
                return point

        start_x, start_y = _handle(start_point)
        end_x, end_y = _handle(end_point)

        version = self.sdk_version
        if version <= 15:
            raise AdbSDKVersionError(f'swipe: API <= 15 not supported (version={version})')
        elif version <= 17:
            self.shell(f'input swipe {start_x} {start_y} {end_x} {end_y} {duration}')
        else:
            self.shell(f'input touchscreen swipe {start_x} {start_y} {end_x} {end_y}')

    def set_input_method(self, ime_method: str, ime_apk_path: Optional[str] = None) -> None:
        """
        设置输入法

        Args:
            ime_method: 输入法ID
            ime_apk_path: 输入法安装包
        Returns:
            None
        """
        if ime_method not in self.ime_list:
            if ime_apk_path:
                self.install(ime_apk_path)
        if self.default_ime != ime_method:
            self.shell(['ime', 'enable', ime_method])
            self.shell(['ime', 'set', ime_method])

    def text(self, text, enter: Optional[bool] = False):
        """
        input text on the device
        预置命令: #CLEAR# 清除当前输入框内所有字符。在使用原生input时不能保证百分百清空输入框数据

        Args:
            text: 需要输入的字符
            enter: press 'Enter' key

        Returns:
            None
        """
        if self.default_ime == ADB_DEFAULT_KEYBOARD:
            if text == '#CLEAR#':
                self.broadcast('ADB_CLEAR_TEXT')
            else:
                self.shell(f"am broadcast -a ADB_INPUT_TEXT --es msg '{str(text)}'")
        else:
            if text == '#CLEAR#':
                logger.warning('建议使用AdbKeyboard')
                for i in range(255):
                    self.keyevent('KEYCODE_CLEAR')
            else:
                self.shell(['input', 'text', str(text)])

        if enter:
            time.sleep(1)
            self.keyevent('ENTER')


class ADBExtraDevice(ADBDevice):
    def __init__(self, aapt: bool = False, minicap: bool = False,
                 *args, **kwargs):
        """
        添加了额外模块的adb

        Args:
            aapt: Android资源打包工具,可以用于解析apk
            minicap: Android流式传输实时屏幕捕获数据
        """
        super(ADBExtraDevice, self).__init__(*args, **kwargs)
        if aapt:
            from adbutils.extra import Aapt
            self.aapt = Aapt(device=self)
            self._aapt_flag = True
        else:
            self._aapt_flag = False

        if minicap:
            from adbutils.extra import Rotation
            self.rotation_watcher = Rotation(device=self)
            self.rotation_watcher.start()

            from adbutils.extra import Minicap
            self.minicap = Minicap(device=self, rotation_watcher=self.rotation_watcher)
            self._minicap_flag = True
        else:
            self._minicap_flag = False

    def __getattribute__(self, item):
        """
        用于隔离自定义模块未创建时的报错

        Args:
            item: 当前访问的属性
        """
        if item == 'minicap' and not self._minicap_flag:
            raise AdbExtraModuleNotFount('minicap not create, Call minicap = True at instance time')
        elif item == 'aapt' and not self._aapt_flag:
            raise AdbExtraModuleNotFount('aapt not create， Call aapt = True at instance time')
        return object.__getattribute__(self, item)


__all__ = ['ADBClient', 'ADBDevice', 'ADBExtraDevice']



# -*- coding: utf-8 -*-
from .aapt import Aapt
from .minicap import Minicap
from .rotation import Rotation
from .performance.fps import Fps
from .performance.cpu import Cpu
from .performance.meminfo import Meminfo
from .performance import DeviceWatcher


__all__ = ['Aapt', 'Minicap', 'Rotation', 'Fps', 'Cpu', 'Meminfo', 'DeviceWatcher']


# -*- coding: utf-8 -*-
import time
import threading

from loguru import logger
from adbutils.constant import (ANDROID_TMP_PATH, MNC_REMOTE_PATH, MNC_SO_REMOTE_PATH, MNC_CMD, MNC_CAP_LOCAL_PATH,
                               MNC_LOCAL_NAME, MNC_LOCAL_PATH, MNC_SO_LOCAL_PATH)
from adbutils.extra.minicap.exceptions import MinicapStartError, MinicapServerConnectError
from adbutils import ADBDevice
from adbutils._utils import NonBlockingStreamReader, reg_cleanup, SafeSocket
from adbutils._wraps import threadsafe_generator

from typing import Tuple
import struct
import subprocess


class Minicap(object):
    RECVTIMEOUT = None

    def __init__(self, device: ADBDevice, rotation_watcher=None):
        """
        初始化minicap

        Args:
            device: 设备类
            rotation_watcher: 方向监控函数
        """
        self.device = device
        self.MNC_LOCAL_NAME = MNC_LOCAL_NAME.format(device_id=self.device.device_id)  # minicap在设备上的转发名
        self.MNC_PORT = None  # minicap在电脑上使用的端口
        self.quirk_flag = 0
        self.server_flag = False  # 判断minicap服务是否启动
        self.proc = None
        self.nbsp = None

        self._update_rotation_event = threading.Event()
        if rotation_watcher:
            rotation_watcher.reg_callback(lambda x: self.update_rotation(x * 90))
        self._install_minicap()

    def __str__(self):
        return f"<minicap ({self.server_flag and 'Start' or 'Close'})> port:{self.MNC_PORT}" \
               f"\tlocal_name:{self.MNC_LOCAL_NAME}"

    def start_server(self) -> None:
        """
        开启minicap服务

        Raises:
            MinicapStartError: minicap server start error
        Returns:
            None
        """
        self._set_minicap_forward()
        param = self._get_params()
        proc = self.device.start_shell([MNC_CMD, f"-n '{self.MNC_LOCAL_NAME}'", '-P',
                                        "%dx%d@%dx%d/%d 2>&1" % param])

        nbsp = NonBlockingStreamReader(proc.stdout)
        while True:
            line = nbsp.readline(timeout=5)
            if line is None:
                raise MinicapStartError("minicap server setup timeout")
            if b'have different types' in line:
                raise MinicapStartError("minicap server setup error")
            if b"Server start" in line:
                logger.info('minicap server setup')
                break

        if proc.poll() is not None:
            raise MinicapStartError('minicap server quit immediately')
        reg_cleanup(proc.kill)
        time.sleep(.5)
        # self.proc = proc
        # self.nbsp = nbsp
        self.server_flag = True

    def teardown(self) -> None:
        """
        关闭minicap服务

        Returns:
            None
        """
        logger.debug('minicap server teardown')
        if self.proc:
            self.proc.kill()

        if self.nbsp:
            self.nbsp.kill()

        if self.MNC_PORT and self.device.get_forward_port(remote=self.MNC_LOCAL_NAME):
            self.device.remove_forward(local=f'tcp:{self.MNC_PORT}')

        self.server_flag = False

    def _set_minicap_forward(self):
        """
        设置minicap开放的端口

        Returns:
            None
        """
        # teardown服务后,保留端口信息,用于下次启动
        remote = f'localabstract:{self.MNC_LOCAL_NAME}'
        if port := self.device.get_forward_port(remote=remote, device_id=self.device.device_id):
            self.MNC_PORT = port
            return

        self.MNC_PORT = self.MNC_PORT or self.device.get_available_forward_local()
        self.device.forward(local=f'tcp:{self.MNC_PORT}', remote=remote)

    def _install_minicap(self) -> None:
        """
        check if minicap and minicap.so installed

        Returns:
            None
        """
        if not self.device.check_file(ANDROID_TMP_PATH, 'minicap'):
            self.device.push(local=MNC_LOCAL_PATH.format(abi_version=self.device.abi_version),
                             remote=MNC_REMOTE_PATH)
            time.sleep(1)
            self.device.shell(['chmod', '755', MNC_REMOTE_PATH])

        if not self.device.check_file(ANDROID_TMP_PATH, 'minicap.so'):
            self.device.push(local=MNC_SO_LOCAL_PATH.format(abi_version=self.device.abi_version,
                                                            sdk_version=self.device.sdk_version),
                             remote=MNC_SO_REMOTE_PATH)
            time.sleep(1)
            self.device.shell(['chmod', '755', MNC_SO_REMOTE_PATH])

    def _get_params(self) -> Tuple[int, int, int, int, int]:
        """
        获取minicap命令需要的屏幕分辨率参数

        Returns:
            None
        """
        display_info = self.device.displayInfo
        real_width = display_info['width']
        real_height = display_info['height']
        real_rotation = display_info['rotation']

        if self.quirk_flag & 2 and real_rotation in (90, 270):
            params = real_height, real_width, real_height, real_width, 0
        else:
            params = real_width, real_height, real_width, real_height, real_rotation

        return params

    def update_rotation(self, rotation):
        """
        更新屏幕方向

        Args:
            rotation: 方向角度

        Returns:
            None
        """
        logger.debug("minicap update_rotation: {}", rotation)
        self._update_rotation_event.set()

    def get_frame(self):
        """
        获取屏幕截图

        Returns:
            图像数据
        """
        if self._update_rotation_event.is_set():
            logger.info('minicap update_rotation')
            self.teardown()
            self.start_server()
            self._update_rotation_event.clear()

        try:
            return self._get_frame()
        except (ConnectionRefusedError, OSError) as err:
            self.teardown()
            raise MinicapServerConnectError(f'{err}')

    def _get_frame(self):
        s = SafeSocket()
        s.connect((self.device.host, self.MNC_PORT))
        t = s.recv(24)
        # minicap header
        global_headers = struct.unpack("<2B5I2B", t)
        # Global header binary format https://github.com/openstf/minicap#global-header-binary-format
        ori, self.quirk_flag = global_headers[-2:]

        if self.quirk_flag & 2 and ori not in (0, 1, 2):
            stopping = True
            logger.error("quirk_flag found:{}, going to resetup", self.quirk_flag)
        else:
            stopping = False

        if not stopping:
            s.send(b"1")
            if self.RECVTIMEOUT is not None:
                header = s.recv_with_timeout(4, self.RECVTIMEOUT)
            else:
                header = s.recv(4)
            if header is None:
                logger.error("minicap header is None")
            else:
                frame_size = struct.unpack("<I", header)[0]
                frame_data = s.recv(frame_size)
                s.close()
                return frame_data

        logger.info('minicap get_frame ends')
        s.close()
        self.teardown()


if __name__ == '__main__':
    import cv2
    from baseImage import IMAGE

    from adbutils import ADBDevice
    from adbutils.extra.minicap import Minicap

    device = ADBDevice(device_id='emulator-5554')
    minicap = Minicap(device)
    minicap.start_server()

    while True:
        if img := minicap.get_frame():
            cv2.imshow('test', IMAGE(img).imread())

        if cv2.waitKey(25) & 0xFF == ord('q'):
            cv2.destroyAllWindows()
            minicap.teardown()



# -*- coding: utf-8 -*-
import time
from threading import Thread, Event
from queue import Queue

from loguru import logger

from adbutils import ADBDevice
from adbutils.exceptions import AdbBaseError
from adbutils.extra.performance.cpu import Cpu
from adbutils.extra.performance.fps import Fps
from adbutils.extra.performance.meminfo import Meminfo

__all__ = ['DeviceWatcher']


class DeviceWatcher(object):
    def __init__(self, device: ADBDevice, package_name: str = None, surfaceView_name: str = None):
        self._surfaceView_name = surfaceView_name
        self._package_name = package_name
        self._device = device

        self._cpu_watcher = Cpu(self._device)
        self._fps_watcher = Fps(self._device)
        self._mem_watcher = Meminfo(self._device)

        self._kill_event = Event()

        self._cpu_usage_queue = Queue()
        self._cpu_wait_event = Event()
        self._cpu_wait_event.set()
        self._cpu_watcher_thread: Thread = self.create_cpu_watcher()

        self._mem_usage_queue = Queue()
        self._mem_wait_event = Event()
        self._mem_wait_event.set()
        self._mem_watcher_thread: Thread = self.create_mem_watcher()

        self._fps_usage_queue = Queue()
        self._fps_wait_event = Event()
        self._fps_wait_event.set()
        self._fps_watcher_thread: Thread = self.create_fps_watcher()

        if not self._surfaceView_name and self._package_name:
            if layers := self._fps_watcher.get_possible_activity():
                surfaceViews = self._fps_watcher.check_activity_usable(layers)
                if self._package_name in surfaceViews:
                    self._surfaceView_name = surfaceViews
            logger.debug(f'自动设置监控activity={self._surfaceView_name}')

    def create_cpu_watcher(self) -> Thread:
        """
        创建cpu监控线程

        Returns:
            cpu监控线程
        """

        def _get_cpu_usage():
            try:
                # total_cpu_usage, cpu_core_usage, app_usage_ret
                return self._cpu_watcher.get_cpu_usage(self._package_name)
            except AdbBaseError as err:
                logger.error(err)
            return None

        def _run(kill_event: Event, wait_event: Event, q: Queue):
            while not kill_event.is_set():
                if not wait_event.is_set():
                    if cpu_usage := _get_cpu_usage():
                        q.put(cpu_usage)
                    else:
                        q.put(None)
                    wait_event.set()

        _t = Thread(target=_run, name='cpu_watcher',
                    args=(self._kill_event, self._cpu_wait_event, self._cpu_usage_queue))
        _t.daemon = True
        return _t

    def create_mem_watcher(self) -> Thread:
        """
        创建内存监控线程

        Returns:
            内存监控线程
        """

        def _get_mem_usage():
            try:
                if self._package_name:
                    return self._mem_watcher.get_app_summary(self._package_name)
                else:
                    return None
            except AdbBaseError as err:
                logger.error(err)
            return None

        def _run(kill_event: Event, wait_event: Event, q: Queue):
            while not kill_event.is_set():
                if not wait_event.is_set():
                    if app_mem := _get_mem_usage():
                        q.put(app_mem)
                    else:
                        q.put(None)
                    wait_event.set()

        _t = Thread(target=_run, name='mem_watcher',
                    args=(self._kill_event, self._mem_wait_event, self._mem_usage_queue))
        _t.daemon = True
        return _t

    def create_fps_watcher(self) -> Thread:
        """
        创建fps监控线程

        Returns:
            fps监控线程
        """

        def _get_fps_usage():
            try:
                if self._surfaceView_name:
                    return self._fps_watcher.get_fps_surfaceView(f"{self._surfaceView_name}")
                return None
            except AdbBaseError as err:
                logger.error(err)
            return None

        def _run(kill_event: Event, wait_event: Event, q: Queue):
            while not kill_event.is_set():
                if not wait_event.is_set():
                    if fps_info := _get_fps_usage():
                        q.put(fps_info)
                    else:
                        q.put(None)
                    wait_event.set()

        _t = Thread(target=_run, name='mem_watcher',
                    args=(self._kill_event, self._fps_wait_event, self._fps_usage_queue))
        _t.daemon = True
        return _t

    def stop(self):
        self._kill_event.set()

    def start(self):
        if not self._cpu_watcher_thread.is_alive():
            self._cpu_watcher_thread.start()

        if not self._mem_watcher_thread.is_alive():
            self._mem_watcher_thread.start()

        if not self._fps_watcher_thread.is_alive():
            self._fps_watcher.clear_surfaceFlinger_latency()
            self._fps_watcher_thread.start()

    def get(self):
        """

        Returns:

        """
        self._mem_wait_event.clear()
        self._cpu_wait_event.clear()
        self._fps_wait_event.clear()

        cpu_usage = self._cpu_usage_queue.get()
        mem_usage = self._mem_usage_queue.get()
        fps_info = self._fps_usage_queue.get()

        return cpu_usage, mem_usage, fps_info


if __name__ == '__main__':
    from adbutils import ADBDevice
    from adbutils.extra.performance import DeviceWatcher

    device = ADBDevice(device_id='')
    a = DeviceWatcher(device, package_name=device.foreground_package)
    a.start()

    while True:
        start_time = time.time()
        cpu_usage, mem_usage, fps_info = a.get()
        delay_time = time.time() - start_time

        log = []
        if cpu_usage:
            total_cpu_usage, cpu_core_usage, app_usage_ret = cpu_usage
            log.append('cpu={} core={}, {}'.format(
                f'{total_cpu_usage:.1f}%',
                '\t'.join([f'cpu{core_index}:{usage:.1f}%' for core_index, usage in enumerate(cpu_core_usage)]),
                '\t'.join([f'{name}:{usage:.1f}%' for name, usage in app_usage_ret.items()]),
            ))

        if fps_info:
            fps, fTime, jank, bigJank, _ = fps_info
        else:
            fps = fTime = 0
        log.append(f'fps={fps:.1f}, 最大延迟={fTime:.2f}ms')

        logger.debug('\t'.join(log))
        if (sleep := (1 - delay_time)) > 0:
            time.sleep(sleep)
        else:
            time.sleep(1)



import platform
import sys
import os
import queue
import socket
import threading
import subprocess
from typing import IO, Optional, Union
from adbutils.constant import DEFAULT_ADB_PATH


def check_file(fileName: str):
    """check file in path"""
    return os.path.isfile(f'{fileName}')


def get_std_encoding(stream):
    """
    Get encoding of the stream
    Args:
        stream: stream
    Returns:
        encoding or file system encoding
    """
    return getattr(stream, "encoding", None) or sys.getfilesystemencoding()


def split_cmd(cmds) -> list:
    """
    Split the commands to the list for subprocess
    Args:
        cmds (str): command
    Returns:
        command list
    """
    # cmds = shlex.split(cmds)  # disable auto removing \ on windows
    return cmds.split() if isinstance(cmds, str) else list(cmds)


def _popen_kwargs() -> dict:
    creationflags = 0
    startupinfo = None
    if sys.platform.startswith('win'):
        try:
            creationflags = subprocess.CREATE_NO_WINDOW  # python 3.7+
        except AttributeError:
            creationflags = 0x8000000
    return {
        'creationflags': creationflags,
        'startupinfo': startupinfo,
    }


def get_adb_exe() -> str:
    """
    获取adb路径
    :return:
    """
    # find in $PATH
    cmds = ['adb', "--version"]
    try:
        with open(os.devnull, "w") as null:
            subprocess.check_call(
                cmds, stdout=null, stderr=subprocess.STDOUT, **_popen_kwargs()
            )
        adb_path = 'adb'
    except (FileNotFoundError, OSError, ValueError):
        system = platform.system()
        machine = platform.machine()
        adb_path = DEFAULT_ADB_PATH.get(f'{system}-{machine}')
        if not adb_path:
            adb_path = DEFAULT_ADB_PATH.get(f'{system}')
        if not adb_path:
            raise RuntimeError(f"No adb executable supports this platform({system}-{machine}).")

    return adb_path


class NonBlockingStreamReader(object):
    # TODO: 增加一个方法用于非阻塞状态将stream输出存入文件
    def __init__(self, stream: IO, raise_EOF: Optional[bool] = False, print_output: bool = True,
                 print_new_line: bool = True):
        self._s = stream
        self._q = queue.Queue()
        self._lastline = None
        self.name = id(self)

        def _populateQueue(_stream: IO, _queue: queue.Queue, kill_event: threading.Event):
            """
            Collect lines from 'stream' and put them in 'queue'

            Args:
                _stream: 文件流
                _queue: 队列
                kill_event: 一个事件管理标志

            Returns:
                None
            """
            while not kill_event.is_set():
                line = _stream.readline()
                if line is not None:
                    _queue.put(line)
                    if print_output:
                        if print_new_line and line == self._lastline:
                            continue
                        self._lastline = line
                elif kill_event.is_set():
                    break
                elif raise_EOF:
                    raise UnexpectedEndOfStream
                else:
                    break

        self._kill_event = threading.Event()
        self._t = threading.Thread(target=_populateQueue, args=(self._s, self._q, self._kill_event))
        self._t.daemon = True
        self._t.start()  # start collecting lines from the stream

    def readline(self, timeout: Union[int] = None):
        try:
            return self._q.get(block=timeout is not None, timeout=timeout)
        except queue.Empty:
            return None

    def read(self) -> bytes:
        lines = []
        while True:
            line = self.readline()
            if line is None:
                break
            lines.append(line)
        return b"".join(lines)

    def kill(self) -> None:
        self._kill_event.set()


class UnexpectedEndOfStream(Exception):
    pass


CLEANUP_CALLS = queue.Queue()


def reg_cleanup(func, *args, **kwargs):
    """
    Clean the register for given function
    Args:
        func: function name
        *args: optional argument
        **kwargs: optional arguments
    Returns:
        None
    """
    CLEANUP_CALLS.put((func, args, kwargs))


class SafeSocket(object):
    """safe and exact recv & send"""
    def __init__(self, sock=None):
        if sock is None:
            self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        else:
            self.sock = sock
        self.buf = b""

    # PEP 3113 -- Removal of Tuple Parameter Unpacking
    # https://www.python.org/dev/peps/pep-3113/
    def connect(self, tuple_hp):
        host, port = tuple_hp
        self.sock.connect((host, port))

    def send(self, msg):
        totalsent = 0
        while totalsent < len(msg):
            sent = self.sock.send(msg[totalsent:])
            if sent == 0:
                raise socket.error("socket connection broken")
            totalsent += sent

    def recv(self, size):
        while len(self.buf) < size:
            trunk = self.sock.recv(min(size-len(self.buf), 4096))
            if trunk == b"":
                raise socket.error("socket connection broken")
            self.buf += trunk
        ret, self.buf = self.buf[:size], self.buf[size:]
        return ret

    def recv_with_timeout(self, size, timeout=2):
        self.sock.settimeout(timeout)
        try:
            ret = self.recv(size)
        except socket.timeout:
            ret = None
        finally:
            self.sock.settimeout(None)
        return ret

    def recv_nonblocking(self, size):
        self.sock.settimeout(0)
        try:
            ret = self.recv(size)
        except socket.error as e:
            # 10035 no data when nonblocking
            if e.args[0] == 10035:  # errno.EWOULDBLOCK
                ret = None
            # 10053 connection abort by client
            # 10054 connection reset by peer
            elif e.args[0] in [10053, 10054]:  # errno.ECONNABORTED:
                raise
            else:
                raise
        return ret

    def close(self):
        self.sock.close()



# -*- coding: utf-8 -*-
import time
import threading
from inspect import isfunction
from functools import wraps
from typing import Optional, Union, Tuple, List, Type


class print_run_time(object):
    def __init__(self):
        pass

    def __call__(self, func):
        @wraps(func)
        def wrapped_function(*args, **kwargs):
            start_time = time.time()
            ret = func(*args, **kwargs)
            print("{}() run time is {time:.2f}ms".format(func.__name__, time=(time.time() - start_time) * 1000))
            return ret
        return wrapped_function


class retries(object):
    def __init__(self, max_tries: int, delay: Optional[int] = 1,
                 exceptions: Tuple[Type[Exception], ...] = (Exception,), hook=None):
        """
        通过装饰器实现的"重试"函数

        Args:
            max_tries: 最大可重试次数。超出次数后仍然失败,则弹出异常
            delay: 重试等待间隔
            exceptions: 需要检测的异常
            hook: 钩子函数
        """
        self.max_tries = max_tries
        self.delay = delay
        self.exceptions = exceptions
        self.hook = hook

    def __call__(self, func):
        @wraps(func)
        def wrapped_function(*args, **kwargs):
            tries = list(range(self.max_tries))
            tries.reverse()
            for tries_remaining in tries:
                try:
                    return func(*args, **kwargs)
                except self.exceptions as err:
                    if tries_remaining > 0:
                        if isfunction(self.hook):
                            self.hook(tries_remaining, err)
                        time.sleep(self.delay)
                    else:
                        raise err
        return wrapped_function


class ThreadSafeIter:
    """
    Takes an iterator/generator and makes it thread-safe by
    serializing call to the `next` method of given iterator/generator.
    """
    def __init__(self, it):
        self.it = it
        self.lock = threading.Lock()
        self._next = self.it.__next__  # py3

    def __iter__(self):
        return self

    def __next__(self):
        with self.lock:
            return self._next()

    def send(self, *args):
        with self.lock:
            return self.it.send(*args)


def threadsafe_generator(func):
    """
    A decorator that takes a generator function and makes it thread-safe.
    """
    @wraps(func)
    def wrapped_function(*a, **kw):
        return ThreadSafeIter(func(*a, **kw))
    return wrapped_function



#! usr/bin/python
# -*- coding:utf-8 -*-
import os
import time
import threading
import traceback
from loguru import logger


class Rotation(object):
    def __init__(self, device):
        self.device = device
        self._kill_event = threading.Event()
        self.current_orientation = None
        self._t = None
        self.ow_callback = []

    def start(self):

        def _refresh_by_adb():
            ori = self.device.getDisplayOrientation()
            return ori

        def _run(kill_event):
            while not kill_event.is_set():
                ori = _refresh_by_adb()
                if self.current_orientation == ori:
                    time.sleep(2)
                    continue
                if ori is None:
                    continue
                logger.info('update orientation {}->{}'.format(self.current_orientation, ori))
                self.current_orientation = ori
                for callback in self.ow_callback:
                    try:
                        callback(ori)
                    except Exception:
                        logger.error('callback: {} error'.format(callback))
                        traceback.print_exc()

        self.current_orientation = _refresh_by_adb()
        self._t = threading.Thread(target=_run, args=(self._kill_event,), name='rotationwatcher')
        self._t.daemon = True
        self._t.start()
        return self.current_orientation

    def reg_callback(self, ow_callback):
        """
        Args:
            ow_callback:
        Returns:
        """
        """方向变化的时候的回调函数，参数一定是ori"""
        self.ow_callback.append(ow_callback)



# -*- coding: utf-8 -*-
from adbutils.exceptions import AdbBaseError


class MinicapStartError(AdbBaseError):
    """ An error while minicap server start error """


class MinicapServerConnectError(AdbBaseError):
    """ An error while minicap server connect error """




# -*- coding: utf-8 -*-
from setuptools import setup

setup(
    name='adb-utils',
    version='1.0.9',
    author='hakaboom',
    author_email='1534225986@qq.com',
    license='Apache License 2.0',
    description='This is a secondary package of adb',
    url='https://github.com/hakaboom/base_image',
    packages=['adbutils'],
    install_requires=["baseImage==1.0.6"],
)


# -*- coding: utf-8 -*-
import re
import time
from typing import Optional, Union, List, Tuple

from adbutils import ADBDevice

from loguru import logger


class Fps(object):
    _MIN_NORMALIZED_FRAME_LENGTH = 0.5
    Movie_FrameTime = 1000 / 24 / 1000  # 电影帧耗时,单位为秒
    nanoseconds_per_second = 1e9  # 纳秒转换成秒
    pending_fence_timestamp = (1 << 63) - 1  # 查询某帧数据出问题的时候，系统会返回一个64位的int最大值，忽略这列数据

    def __init__(self, device: ADBDevice):
        self.device = device
        self._last_drawEnd_timestamps = None

    def get_fps_surfaceView(self, surface_name: str):
        _Fps = 0
        _FTime = 0
        _Jank = 0
        _BigJank = 0
        _Stutter = 0

        # step1: 根据window名,获取帧数信息
        stat = self._get_surfaceFlinger_stat(surface_name)

        # step2: 提取帧数信息,分别得到刷新周期/绘制图像开始时间列表/绘制耗时列表/绘制结束列表
        refresh_period, _drawStart_timestamps, _vsync_timestamps, _drawEnd_timestamps = \
            self._pares_surfaceFlinger_stat(stat)

        if not _drawStart_timestamps or not _vsync_timestamps or not _drawEnd_timestamps:
            return None

        # step3: 根据上次获取的最后一帧时间戳,切分本次有效帧列表
        if self._last_drawEnd_timestamps in _drawEnd_timestamps:
            index = list(reversed(_drawEnd_timestamps)).index(self._last_drawEnd_timestamps)
            index = len(_drawEnd_timestamps) - index + 1
            if index == len(_drawEnd_timestamps) - 1:
                index = 0
        else:
            index = 0

        drawStart_timestamps = _drawStart_timestamps[index:]
        vsync_timestamps = _vsync_timestamps[index:]
        drawEnd_timestamps = _drawEnd_timestamps[index:]

        # step4: 计算FPS(帧数)
        if (frame_count := len(vsync_timestamps)) < 2:
            return None
        seconds = vsync_timestamps[-1] - vsync_timestamps[0]
        _Fps = round((frame_count - 1) / seconds, 2)

        # step5: 计算FTime(帧耗时),
        vsync_frameTimes = self._get_frameTimes(vsync_timestamps)
        if vsync_frameTimes:
            _FTime = max(vsync_frameTimes)
            _FTime *= 1000

        # step6: 计算Jank
        # 由于step3只截取了有效帧,因此需要额外收集前三帧,用于计算帧耗时
        if index - 3 >= 0:
            jank_vsync_timestamps = _vsync_timestamps[index - 3:]
        else:
            jank_vsync_timestamps = vsync_timestamps
        jank_vsync_frameTimes = self._get_frameTimes(jank_vsync_timestamps)
        _Jank, _BigJank, Jank_time = self._get_perfdog_jank(jank_vsync_frameTimes)
        _Stutter = Jank_time / seconds * 100

        # step End: 记录最后一帧时间戳
        self._last_drawEnd_timestamps = drawEnd_timestamps[-1]

        return _Fps, _FTime, _Jank, _BigJank, _Stutter

    def clear_surfaceFlinger_latency(self) -> bool:
        """
        command 'adb shell dumpsys SurfaceFlinger --latency-clear' 清除SurfaceFlinger latency里的数据

        Returns:
            True设备支持dumpsys SurfaceFlinger --latency, 否则为False
        """
        ret = self.device.shell(['dumpsys', 'SurfaceFlinger', '--latency-clear'])
        return not len(ret)

    def _get_surfaceFlinger_stat(self, surface_name: Optional[str] = None) -> str:
        """
        command 'adb shell dumpsys SurfaceFlinger --latency <Surface Name>

        Returns:
            sufaceFlinger stat
        """
        ret = self.device.shell(['dumpsys', 'SurfaceFlinger', '--latency', self._pares_activity_name(surface_name)])
        if len(ret.splitlines()) > 1:
            return ret
        surface_name = self.get_possible_activity() or surface_name
        logger.warning('warning to get surfaceFlinger try again')
        return self._get_surfaceFlinger_stat(surface_name)

    def _pares_surfaceFlinger_stat(self, stat: str) -> Tuple[float, List[float], List[float], List[float]]:
        """
        处理SurfaceFlinger的信息，返回(刷新周期/绘制图像开始时间列表/绘制耗时列表/绘制结束列表)

        like:
            16666666
            10771456257842  10771499569140	10771456257842
            10771473192729	10771516235806	10771473192729
            10771490277889	10771532902472	10771490277889
            10771507357378	10771549569138	10771507357378
            10771523229435	10771566235804	10771523229435
            ...

        Args:
            stat: dumpsys SurfaceFlinger获得的信息

        Returns:
            刷新周期/绘制图像开始时间列表/绘制耗时列表/绘制结束列表
        """
        stat = stat.splitlines()
        # 三个列表分别对应dumpsys SurfaceFlinger的每一列

        # A) when the app started to draw
        # 开始绘制图像的瞬时时间
        drawStart_timestamps = []

        # B) the vsync immediately preceding SF submitting the frame to the h/w
        # 垂直同步软件把帧提交给硬件之前的瞬时时间戳;VSYNC信令将软件SF帧传递给硬件HW之前的垂直同步时间
        vsync_timestamps = []

        # C) timestamp immediately after SF submitted that frame to the h/w
        # 完成绘制的瞬时时间;SF将帧传递给HW的瞬时时间，
        drawEnd_timestamps = []

        # 刷新周期
        refresh_period = int(stat[0]) / self.nanoseconds_per_second

        # 清除无用的空数据
        empty_data = [0, 0, 0]
        pattern = re.compile(r'(\d+)\s*(\d+)\s*(\d+)')
        for line in stat[1:]:
            # 确认数据结构,与数据是否有效
            fields = pattern.search(line)
            if not fields:
                continue
            
            # 判断数据是否有效
            fields = [int(v) for v in fields.groups()]
            if fields == empty_data or self.pending_fence_timestamp in fields:
                continue

            drawStart_timestamp = fields[0]
            drawStart_timestamp /= self.nanoseconds_per_second

            vsync_timestamp = fields[1]
            vsync_timestamp /= self.nanoseconds_per_second

            drawEnd_timestamp = fields[2]
            drawEnd_timestamp /= self.nanoseconds_per_second

            drawStart_timestamps.append(drawStart_timestamp)
            vsync_timestamps.append(vsync_timestamp)
            drawEnd_timestamps.append(drawEnd_timestamp)

        return refresh_period, drawStart_timestamps, vsync_timestamps, drawEnd_timestamps

    @staticmethod
    def _get_frameTimes(data: List[float]) -> List[float]:
        """
        计算两帧渲染耗时

        Args:
            data: 包含帧时间戳的列表

        Returns:
            两帧渲染耗时列表
        """
        deltas = [t2 - t1 for t1, t2 in zip(data, data[1:])]
        return deltas

    def _get_perfdog_jank(self, data: List[float]) -> Tuple[int, int, float]:
        """
        根据每帧耗时,计算jank

        同时满足两条件，则认为是一次卡顿Jank.
            1. FrameTime>前三帧平均耗时2倍。
            2. FrameTime>两帧电影帧耗时 (1000ms/24*2≈83.33ms)。

        同时满足两条件，则认为是一次严重卡顿BigJank.
            1. FrameTime >前三帧平均耗时2倍。
            2. FrameTime >三帧电影帧耗时(1000ms/24*3=125ms)
        Args:
            data: 每帧渲染耗时

        Returns:

        """
        _jank = [new for new, *last in zip(data[3:], data[2:], data[1:], data)
                 if (new >= self.Movie_FrameTime * 2) and (new > (sum(last) / 3) * 2)]

        _bigJank = [new for new, *last in zip(data[3:], data[2:], data[1:], data)
                    if (new >= self.Movie_FrameTime * 3) and (new > (sum(last) / 3) * 2)]

        jank = len(_jank)
        bigJank = len(_bigJank)
        jank_time = sum(_jank) + sum(_bigJank)

        return jank, bigJank, jank_time

    def get_possible_activity(self) -> Optional[str]:
        """
        通过 ‘dumpsys SurfaceFlinger --list',查找到当前最顶部层级名

        Returns:
            包含可能层级
        """
        ret = self.device.shell(['dumpsys', 'SurfaceFlinger', '--list']).strip().splitlines()
        # 特殊适配谷歌手机
        if self.device.manufacturer == 'Google':
            # adb shell dumpsys SurfaceFlinger --latency 'SurfaceView[xx.xx.xx.xx/org.xx.lua.AppActivity](BLAST)#0'
            buffering_stats_pattern = re.compile(r'SurfaceView\[.*\(BLAST\).*', re.DOTALL)
        else:
            buffering_stats_pattern = re.compile(r'SurfaceView -.*', re.DOTALL)
        for layer in ret:
            if layers := buffering_stats_pattern.search(layer):
                return layers.group()
        logger.error("Don't find SurfaceView")

    def check_activity_usable(self, activity_name: str) -> Optional[str]:
        """
        检查activity是否有效,command 'adb shell dumpsys SurfaceFlinge --latency <activity>'
        如果只返回了刷新周期,则认为该activity无效

        Args:
            activity_name: 需要检查的activity名

        Returns:
            满足条件的activity
        """
        if stat := self._get_surfaceFlinger_stat(activity_name):
            stat_len = len(stat.splitlines())
            if stat_len > 1:
                return activity_name

    @staticmethod
    def _pares_activity_name(activity_name: str = None) -> Optional[str]:
        """
        检查activity名是否符合标准

        Args:
            activity_name: activity名

        Returns:
            处理后的activity名
        """
        if not activity_name or not isinstance(activity_name, str):
            return ''

        # 检查是否使用冒号包裹
        pattern = re.compile(r"^'(.*)'$")
        if pattern.search(activity_name):
            return activity_name
        # 包含空格和小括号都要被冒号包裹
        pattern = re.compile(r'\s|\(')
        if pattern.search(activity_name):
            activity_name = f"'{activity_name}'"
        return activity_name



# -*- coding: utf-8 -*-
import os
import re

IP_PATTERN = re.compile(r'(\d+\.){3}\d+')
THISPATH = os.path.dirname(os.path.realpath(__file__))
STATICPATH = os.path.join(THISPATH, 'static')
DEFAULT_ADB_PATH = {
    "Windows": os.path.join(STATICPATH, "adb", "windows", 'adb'),
    "Darwin": os.path.join(STATICPATH, "adb", "mac", "adb"),
    "Linux": os.path.join(STATICPATH, "adb", "linux", "adb"),
    "Linux-x86_64": os.path.join(STATICPATH, "adb", "linux", "adb"),
    "Linux-armv7l": os.path.join(STATICPATH, "adb", "linux_arm", "adb"),
}
ANDROID_ADB_SERVER_HOST = '127.0.0.1'
ANDROID_ADB_SERVER_PORT = 5037

ANDROID_TMP_PATH = '/data/local/tmp/'
ADB_CAP_RAW_REMOTE_PATH = os.path.join(ANDROID_TMP_PATH, 'screencap.raw')

ADB_CAP_RAW_LOCAL_PATH = './{device_id}.raw'
ADB_CAP_LOCAL_PATH = './{device_id}.png'

ADB_DEFAULT_KEYBOARD = 'com.android.adbkeyboard/.AdbIME'
ADB_KEYBOARD_APK_PATH = os.path.join(STATICPATH, 'ADBKeyboard.apk')

ADB_INSTALL_FAILED = {
    "INSTALL_FAILED_ALREADY_EXISTS": "应用已经存在，或卸载了但没卸载干净;建议使用'-r'安装",
    "INSTALL_FAILED_INVALID_APK": "无效的 APK 文件",
    "INSTALL_FAILED_INVALID_URI": "无效的 APK 文件名;确保 APK 文件名里无中文",
    "INSTALL_FAILED_INSUFFICIENT_STORAGE": "空间不足",
    "INSTALL_FAILED_DUPLICATE_PACKAGE": "已经存在同名程序",
    "INSTALL_FAILED_NO_SHARED_USER": "请求的共享用户不存在",
    "INSTALL_FAILED_UPDATE_INCOMPATIBLE": "以前安装过同名应用，但卸载时数据没有移除；或者已安装该应用，但签名不一致;先 adb uninstall",
    "INSTALL_FAILED_SHARED_USER_INCOMPATIBLE": "请求的共享用户存在但签名不一致",
    "INSTALL_FAILED_MISSING_SHARED_LIBRARY": "安装包使用了设备上不可用的共享库",
    "INSTALL_FAILED_REPLACE_COULDNT_DELETE": "替换时无法删除",
    "INSTALL_FAILED_DEXOPT": "dex 优化验证失败或空间不足",
    "INSTALL_FAILED_OLDER_SDK": "设备系统版本低于应用要求",
    "INSTALL_FAILED_CONFLICTING_PROVIDER": "设备里已经存在与应用里同名的 content provider",
    "INSTALL_FAILED_NEWER_SDK": "设备系统版本高于应用要求",
    "INSTALL_FAILED_TEST_ONLY": "应用是 test-only 的，但安装时没有指定 -t 参数",
    "INSTALL_FAILED_CPU_ABI_INCOMPATIBLE": "包含不兼容设备 CPU 应用程序二进制接口的 native code",
    "INSTALL_FAILED_MISSING_FEATURE": "应用使用了设备不可用的功能",
    "INSTALL_FAILED_CONTAINER_ERROR": "1. sdcard 访问失败;2. 应用签名与 ROM 签名一致，被当作内置应用。",
    "INSTALL_FAILED_INVALID_INSTALL_LOCATION": "1. 不能安装到指定位置;2. 应用签名与 ROM 签名一致，被当作内置应用。",
    "INSTALL_FAILED_MEDIA_UNAVAILABLE": "安装位置不可用",
    "INSTALL_FAILED_VERIFICATION_TIMEOUT": "验证安装包超时",
    "INSTALL_FAILED_VERIFICATION_FAILURE": "验证安装包失败",
    "INSTALL_FAILED_PACKAGE_CHANGED": "应用与调用程序期望的不一致",
    "INSTALL_FAILED_UID_CHANGED": "以前安装过该应用，与本次分配的 UID 不一致",
    "INSTALL_FAILED_VERSION_DOWNGRADE": "已经安装了该应用更高版本",
    "INSTALL_FAILED_PERMISSION_MODEL_DOWNGRADE": "已安装 target SDK 支持运行时权限的同名应用，要安装的版本不支持运行时权限",
    "INSTALL_PARSE_FAILED_NOT_APK": "指定路径不是文件，或不是以 .apk 结尾",
    "INSTALL_PARSE_FAILED_BAD_MANIFEST": "无法解析的 AndroidManifest.xml 文件",
    "INSTALL_PARSE_FAILED_UNEXPECTED_EXCEPTION": "解析器遇到异常",
    "INSTALL_PARSE_FAILED_NO_CERTIFICATES": "安装包没有签名",
    "INSTALL_PARSE_FAILED_INCONSISTENT_CERTIFICATES": "已安装该应用，且签名与 APK 文件不一致",
    "INSTALL_PARSE_FAILED_CERTIFICATE_ENCODING": "解析 APK 文件时遇到 CertificateEncodingException",
    "INSTALL_PARSE_FAILED_BAD_PACKAGE_NAME": "manifest 文件里没有或者使用了无效的包名",
    "INSTALL_PARSE_FAILED_BAD_SHARED_USER_ID": "manifest 文件里指定了无效的共享用户 ID",
    "INSTALL_PARSE_FAILED_MANIFEST_MALFORMED": "解析 manifest 文件时遇到结构性错误",
    "INSTALL_PARSE_FAILED_MANIFEST_EMPTY": "在 manifest 文件里找不到找可操作标签（instrumentation 或 application）",
    "INSTALL_FAILED_INTERNAL_ERROR": "因系统问题安装失败",
    "INSTALL_FAILED_USER_RESTRICTED": "用户被限制安装应用",
    "INSTALL_FAILED_DUPLICATE_PERMISSION": "应用尝试定义一个已经存在的权限名称",
    "INSTALL_FAILED_NO_MATCHING_ABIS": "应用包含设备的应用程序二进制接口不支持的 native code",
    "INSTALL_CANCELED_BY_USER": "应用安装需要在设备上确认，但未操作设备或点了取消",
    "INSTALL_FAILED_ACWF_INCOMPATIBLE": "应用程序与设备不兼容",
    "does not contain AndroidManifest.xml": "无效的 APK 文件",
    "is not a valid zip file": "无效的 APK 文件",
    "Offline": "设备未连接成功",
    "unauthorized": "设备未授权允许调试",
    "error: device not found": "没有连接成功的设备",
    "protocol failure": "设备已断开连接",
    "Unknown option: -s": "Android 2.2 以下不支持安装到 sdcard",
    "No space left on device": "空间不足",
    "Permission denied … sdcard …": "sdcard 不可用",
    "signatures do not match the previously installed version; ignoring!": "已安装该应用且签名不一致",
}

# AAPT
AAPT_LOCAL_PATH = {
    'arm': os.path.join(STATICPATH, 'aapt', 'aapt-arm-pie'),
    'x86': os.path.join(STATICPATH, 'aapt', 'aapt-x86-pie'),
}
AAPT_REMOTE_PATH = os.path.join(ANDROID_TMP_PATH, 'aapt')


# busyBox
BUSYBOX_LOCAL_PATH = os.path.join(STATICPATH, 'busybox', 'busybox-arm{}')
BUSYBOX_REMOTE_PATH = os.path.join(ANDROID_TMP_PATH, 'busybox')


# minicap
MNC_REMOTE_PATH = os.path.join(ANDROID_TMP_PATH, 'minicap')
MNC_SO_REMOTE_PATH = os.path.join(ANDROID_TMP_PATH, 'minicap.so')
MNC_CMD = f'LD_LIBRARY_PATH={ANDROID_TMP_PATH} {MNC_REMOTE_PATH}'
MNC_CAP_LOCAL_PATH = ADB_CAP_LOCAL_PATH
MNC_LOCAL_NAME = 'minicap_{device_id}'
MNC_LOCAL_PATH = os.path.join(STATICPATH, 'stf_libs', '{abi_version}', 'minicap')
MNC_SO_LOCAL_PATH = os.path.join(STATICPATH, 'stf_libs', 'minicap-shared', 'aosp', 'libs',
                                   'android-{sdk_version}', '{abi_version}', 'minicap.so')

# minitouch
MNT_REMOTE_PATH = os.path.join(ANDROID_TMP_PATH, 'minitouch')
MNT_LOCAL_NAME = 'minitouch_{device_id}'
MNT_LOCAL_PATH = os.path.join(STATICPATH, 'stf_libs', '{abi_version}', 'minitouch')
